"""ProtocolEngine-based Protocol API core implementation."""

from __future__ import annotations
from typing import Dict, Optional, Type, Union, List, Tuple, TYPE_CHECKING

from opentrons_shared_data.liquid_classes import LiquidClassDefinitionDoesNotExist
from opentrons_shared_data.deck.types import DeckDefinitionV5, SlotDefV3
from opentrons_shared_data.labware.labware_definition import (
    labware_definition_type_adapter,
)
from opentrons_shared_data.labware.types import LabwareDefinition as LabwareDefDict
from opentrons_shared_data import liquid_classes
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassSchemaV1,
)
from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons_shared_data.robot.types import RobotType

from opentrons.types import (
    DeckSlotName,
    Location,
    Mount,
    MountType,
    Point,
    StagingSlotName,
)
from opentrons.hardware_control import SyncHardwareAPI, SynchronousAdapter
from opentrons.hardware_control.modules import AbstractModule
from opentrons.hardware_control.modules.types import ModuleModel, ModuleType
from opentrons.hardware_control.types import DoorState
from opentrons.protocols.api_support.util import AxisMaxSpeeds
from opentrons.protocols.api_support.types import APIVersion

from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.commands import LoadModuleResult
from opentrons.protocol_engine import (
    DeckSlotLocation,
    AddressableAreaLocation,
    ModuleLocation,
    OnLabwareLocation,
    ModuleModel as EngineModuleModel,
    LabwareMovementStrategy,
    LabwareOffsetVector,
    LoadedLabware,
    LoadedModule,
)
from opentrons.protocol_engine.types import (
    ModuleModel as ProtocolEngineModuleModel,
    OFF_DECK_LOCATION,
    SYSTEM_LOCATION,
    LoadableLabwareLocation,
    NonStackedLocation,
)
from opentrons.protocol_engine.clients import SyncClient as ProtocolEngineClient
from opentrons.protocol_engine.errors import (
    LabwareNotLoadedOnModuleError,
    LabwareNotLoadedOnLabwareError,
)

from ... import validation
from ..._types import OffDeckType
from ..._liquid import Liquid, LiquidClass
from ...disposal_locations import TrashBin, WasteChute
from ..protocol import AbstractProtocol
from ..labware import LabwareLoadParams
from .labware import LabwareCore
from .instrument import InstrumentCore
from .robot import RobotCore
from .module_core import (
    ModuleCore,
    TemperatureModuleCore,
    MagneticModuleCore,
    ThermocyclerModuleCore,
    HeaterShakerModuleCore,
    NonConnectedModuleCore,
    MagneticBlockCore,
    AbsorbanceReaderCore,
    FlexStackerCore,
)
from .exceptions import InvalidModuleLocationError
from . import load_labware_params, deck_conflict, overlap_versions
from opentrons.protocol_engine.resources import labware_validation

if TYPE_CHECKING:
    from ...labware import Labware


class ProtocolCore(
    AbstractProtocol[
        InstrumentCore,
        LabwareCore,
        Union[ModuleCore, NonConnectedModuleCore],
    ]
):
    """Protocol API core using a ProtocolEngine.

    Args:
        engine_client: A client to the ProtocolEngine that is executing the protocol.
        api_version: The Python Protocol API versionat which  this core is operating.
        sync_hardware: A SynchronousAdapter-wrapped Hardware Control API.
    """

    def __init__(
        self,
        engine_client: ProtocolEngineClient,
        api_version: APIVersion,
        sync_hardware: SyncHardwareAPI,
    ) -> None:
        self._engine_client = engine_client
        self._api_version = api_version
        self._sync_hardware = sync_hardware
        self._last_location: Optional[Union[Location, TrashBin, WasteChute]] = None
        self._last_mount: Optional[Mount] = None
        self._labware_cores_by_id: Dict[str, LabwareCore] = {}
        self._module_cores_by_id: Dict[
            str, Union[ModuleCore, NonConnectedModuleCore]
        ] = {}
        self._disposal_locations: List[Union[Labware, TrashBin, WasteChute]] = []
        self._liquid_class_def_cache: Dict[Tuple[str, int], LiquidClassSchemaV1] = {}
        self._load_fixed_trash()

    @property
    def api_version(self) -> APIVersion:
        """Get the api version protocol target."""
        return self._api_version

    @property
    def robot_type(self) -> RobotType:
        return self._engine_client.state.config.robot_type

    @property
    def fixed_trash(self) -> Optional[LabwareCore]:
        """Get the fixed trash labware."""
        trash_id = self._engine_client.state.labware.get_fixed_trash_id()
        if trash_id is not None and self._api_version < APIVersion(2, 16):
            return self._labware_cores_by_id[trash_id]
        return None

    def _load_fixed_trash(self) -> None:
        if self.robot_type == "OT-2 Standard" or self._api_version < APIVersion(2, 16):
            trash_id = self._engine_client.state.labware.get_fixed_trash_id()
            if trash_id is not None:
                self._labware_cores_by_id[trash_id] = LabwareCore(
                    labware_id=trash_id,
                    engine_client=self._engine_client,
                )

    def append_disposal_location(
        self,
        disposal_location: Union[Labware, TrashBin, WasteChute],
    ) -> None:
        """Append a disposal location object to the core."""
        self._disposal_locations.append(disposal_location)

    def _add_disposal_location_to_engine(
        self, disposal_location: Union[TrashBin, WasteChute]
    ) -> None:
        """Verify and add disposal location to engine store and append it to the core."""
        self._engine_client.state.addressable_areas.raise_if_area_not_in_deck_configuration(
            disposal_location.area_name
        )
        if isinstance(disposal_location, TrashBin):
            deck_conflict.check(
                engine_state=self._engine_client.state,
                new_trash_bin=disposal_location,
                existing_disposal_locations=self._disposal_locations,
                # TODO: We can now fetch these IDs from engine too.
                #  See comment in self.load_labware().
                #
                # Wrapping .keys() in list() is just to make Decoy verification easier.
                existing_labware_ids=list(self._labware_cores_by_id.keys()),
                existing_module_ids=list(self._module_cores_by_id.keys()),
            )
        self._engine_client.add_addressable_area(disposal_location.area_name)
        self.append_disposal_location(disposal_location)

    def get_disposal_locations(self) -> List[Union[Labware, TrashBin, WasteChute]]:
        """Get disposal locations."""
        return self._disposal_locations

    def get_max_speeds(self) -> AxisMaxSpeeds:
        """Get a control interface for maximum move speeds."""
        raise NotImplementedError("ProtocolCore.get_max_speeds not implemented")

    def get_hardware(self) -> SyncHardwareAPI:
        """Get direct access to a hardware control interface."""
        return self._sync_hardware

    def is_simulating(self) -> bool:
        """Get whether the protocol is being analyzed or actually run."""
        return self._engine_client.state.config.ignore_pause

    def add_labware_definition(
        self,
        definition: LabwareDefDict,
    ) -> LabwareLoadParams:
        """Add a labware definition to the set of loadable definitions."""
        uri = self._engine_client.add_labware_definition(
            labware_definition_type_adapter.validate_python(definition)
        )
        return LabwareLoadParams.from_uri(uri)

    def load_labware(
        self,
        load_name: str,
        location: Union[
            DeckSlotName,
            StagingSlotName,
            LabwareCore,
            ModuleCore,
            NonConnectedModuleCore,
            OffDeckType,
        ],
        label: Optional[str],
        namespace: Optional[str],
        version: Optional[int],
    ) -> LabwareCore:
        """Load a labware using its identifying parameters."""
        load_location = self._convert_labware_location(location=location)

        custom_labware_params = (
            self._engine_client.state.labware.find_custom_labware_load_params()
        )
        namespace, version = load_labware_params.resolve(
            load_name, namespace, version, custom_labware_params, self._api_version
        )

        load_result = self._engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                loadName=load_name,
                location=load_location,
                namespace=namespace,
                version=version,
                displayName=label,
            )
        )
        # FIXME(jbl, 2023-08-14) validating after loading the object issue
        validation.ensure_definition_is_labware(load_result.definition)
        validation.ensure_definition_is_not_lid_after_api_version(
            self.api_version, load_result.definition
        )

        # FIXME(mm, 2023-02-21):
        #
        # We're wrongly checking for deck conflicts *after* we've already loaded the
        # labware into the ProtocolEngine. If it turns out there is a conflict,
        # and this check raises, it will leave this object and its ProtocolEngine
        # in a confusing inconsistent state.
        #
        # I expect we can get away with this in practice a lot of the time because
        # exceptions in Python protocols are mostly treated as fatal, anyway.
        # Users rarely catch them.
        deck_conflict.check(
            engine_state=self._engine_client.state,
            new_labware_id=load_result.labwareId,
            existing_disposal_locations=self._disposal_locations,
            # TODO (spp, 2023-11-27): We've been using IDs from _labware_cores_by_id
            #  and _module_cores_by_id instead of getting the lists directly from engine
            #  because of the chance of engine carrying labware IDs from LPC too.
            #  But with https://github.com/Opentrons/opentrons/pull/13943,
            #  & LPC in maintenance runs, we can now rely on engine state for these IDs too.
            # Wrapping .keys() in list() is just to make Decoy verification easier.
            existing_labware_ids=list(self._labware_cores_by_id.keys()),
            existing_module_ids=list(self._module_cores_by_id.keys()),
        )

        labware_core = LabwareCore(
            labware_id=load_result.labwareId,
            engine_client=self._engine_client,
        )

        self._labware_cores_by_id[labware_core.labware_id] = labware_core

        return labware_core

    def load_adapter(
        self,
        load_name: str,
        location: Union[
            DeckSlotName,
            StagingSlotName,
            ModuleCore,
            NonConnectedModuleCore,
            OffDeckType,
        ],
        namespace: Optional[str],
        version: Optional[int],
    ) -> LabwareCore:
        """Load an adapter using its identifying parameters"""
        load_location = self._get_non_stacked_location(location=location)

        custom_labware_params = (
            self._engine_client.state.labware.find_custom_labware_load_params()
        )
        namespace, version = load_labware_params.resolve(
            load_name, namespace, version, custom_labware_params, self._api_version
        )
        load_result = self._engine_client.execute_command_without_recovery(
            cmd.LoadLabwareParams(
                loadName=load_name,
                location=load_location,
                namespace=namespace,
                version=version,
            )
        )
        # FIXME(jbl, 2023-08-14) validating after loading the object issue
        validation.ensure_definition_is_adapter(load_result.definition)

        # FIXME(jbl, 2023-06-23) read fixme above:
        deck_conflict.check(
            engine_state=self._engine_client.state,
            new_labware_id=load_result.labwareId,
            existing_disposal_locations=self._disposal_locations,
            # TODO: We can now fetch these IDs from engine too.
            #  See comment in self.load_labware().
            #
            # Wrapping .keys() in list() is just to make Decoy verification easier.
            existing_labware_ids=list(self._labware_cores_by_id.keys()),
            existing_module_ids=list(self._module_cores_by_id.keys()),
        )

        labware_core = LabwareCore(
            labware_id=load_result.labwareId,
            engine_client=self._engine_client,
        )

        self._labware_cores_by_id[labware_core.labware_id] = labware_core

        return labware_core

    def load_lid(
        self,
        load_name: str,
        location: LabwareCore,
        namespace: Optional[str],
        version: Optional[int],
    ) -> LabwareCore:
        """Load an individual lid using its identifying parameters. Must be loaded on an existing Labware."""
        load_location = self._convert_labware_location(location=location)
        custom_labware_params = (
            self._engine_client.state.labware.find_custom_labware_load_params()
        )
        namespace, version = load_labware_params.resolve(
            load_name, namespace, version, custom_labware_params, self._api_version
        )
        load_result = self._engine_client.execute_command_without_recovery(
            cmd.LoadLidParams(
                loadName=load_name,
                location=load_location,
                namespace=namespace,
                version=version,
            )
        )
        # FIXME(chb, 2024-12-06) validating after loading the object issue
        validation.ensure_definition_is_lid(load_result.definition)

        deck_conflict.check(
            engine_state=self._engine_client.state,
            new_labware_id=load_result.labwareId,
            existing_disposal_locations=self._disposal_locations,
            # TODO: We can now fetch these IDs from engine too.
            #  See comment in self.load_labware().
            #
            # Wrapping .keys() in list() is just to make Decoy verification easier.
            existing_labware_ids=list(self._labware_cores_by_id.keys()),
            existing_module_ids=list(self._module_cores_by_id.keys()),
        )

        labware_core = LabwareCore(
            labware_id=load_result.labwareId,
            engine_client=self._engine_client,
        )

        self._labware_cores_by_id[labware_core.labware_id] = labware_core
        return labware_core

    def load_labware_to_flex_stacker_hopper(
        self,
        module_core: Union[ModuleCore, NonConnectedModuleCore],
        load_name: str,
        quantity: int,
        label: Optional[str],
        namespace: Optional[str],
        version: Optional[int],
        lid: Optional[str],
    ) -> None:
        """Load one or more labware with or without a lid to the flex stacker hopper."""
        assert isinstance(module_core, FlexStackerCore)
        for _ in range(quantity):
            labware_core = self.load_labware(
                load_name=load_name,
                location=module_core,
                label=label,
                namespace=namespace,
                version=version,
            )
            if lid is not None:
                self.load_lid(
                    load_name=lid,
                    location=labware_core,
                    namespace=namespace,
                    version=version,
                )

    def move_labware(
        self,
        labware_core: LabwareCore,
        new_location: Union[
            DeckSlotName,
            StagingSlotName,
            LabwareCore,
            ModuleCore,
            NonConnectedModuleCore,
            OffDeckType,
            WasteChute,
            TrashBin,
        ],
        use_gripper: bool,
        pause_for_manual_move: bool,
        pick_up_offset: Optional[Tuple[float, float, float]],
        drop_offset: Optional[Tuple[float, float, float]],
    ) -> None:
        """Move the given labware to a new location."""
        if use_gripper:
            strategy = LabwareMovementStrategy.USING_GRIPPER
        elif pause_for_manual_move:
            strategy = LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE
        else:
            strategy = LabwareMovementStrategy.MANUAL_MOVE_WITHOUT_PAUSE

        _pick_up_offset = (
            LabwareOffsetVector(
                x=pick_up_offset[0], y=pick_up_offset[1], z=pick_up_offset[2]
            )
            if pick_up_offset
            else None
        )
        _drop_offset = (
            LabwareOffsetVector(x=drop_offset[0], y=drop_offset[1], z=drop_offset[2])
            if drop_offset
            else None
        )

        to_location = self._convert_labware_location(location=new_location)

        self._engine_client.execute_command(
            cmd.MoveLabwareParams(
                labwareId=labware_core.labware_id,
                newLocation=to_location,
                strategy=strategy,
                pickUpOffset=_pick_up_offset,
                dropOffset=_drop_offset,
            )
        )

        if strategy == LabwareMovementStrategy.USING_GRIPPER:
            # Clear out last location since it is not relevant to pipetting
            # and we only use last location for in-place pipetting commands
            self.set_last_location(location=None, mount=Mount.EXTENSION)

        # FIXME(jbl, 2024-01-04) deck conflict after execution logic issue, read notes in load_labware for more info:
        deck_conflict.check(
            engine_state=self._engine_client.state,
            new_labware_id=labware_core.labware_id,
            existing_disposal_locations=self._disposal_locations,
            # TODO: We can now fetch these IDs from engine too.
            #  See comment in self.load_labware().
            existing_labware_ids=[
                labware_id
                for labware_id in self._labware_cores_by_id
                if labware_id != labware_core.labware_id
            ],
            existing_module_ids=list(self._module_cores_by_id.keys()),
        )

    def move_lid(  # noqa: C901
        self,
        source_location: Union[DeckSlotName, StagingSlotName, LabwareCore],
        new_location: Union[
            DeckSlotName,
            StagingSlotName,
            LabwareCore,
            OffDeckType,
            WasteChute,
            TrashBin,
        ],
        use_gripper: bool,
        pause_for_manual_move: bool,
        pick_up_offset: Optional[Tuple[float, float, float]],
        drop_offset: Optional[Tuple[float, float, float]],
    ) -> LabwareCore | None:
        """Move the given lid to a new location."""
        if use_gripper:
            strategy = LabwareMovementStrategy.USING_GRIPPER
        elif pause_for_manual_move:
            strategy = LabwareMovementStrategy.MANUAL_MOVE_WITH_PAUSE
        else:
            strategy = LabwareMovementStrategy.MANUAL_MOVE_WITHOUT_PAUSE

        if isinstance(source_location, DeckSlotName) or isinstance(
            source_location, StagingSlotName
        ):
            # Find the source labware at the provided deck slot
            labware_in_slot = self._engine_client.state.labware.get_by_slot(
                source_location
            )
            if labware_in_slot is None:
                raise LabwareNotLoadedOnLabwareError(
                    "Lid cannot be loaded on non-labware position."
                )
            else:
                labware = LabwareCore(labware_in_slot.id, self._engine_client)
        else:
            labware = source_location

        # if this is a labware stack, we need to find the labware at the top of the stack
        if labware_validation.is_lid_stack(labware.load_name):
            lid_id = self._engine_client.state.labware.get_highest_child_labware(
                labware.labware_id
            )
        # if this is a labware with a lid, we just need to find its lid_id
        else:
            lid = self._engine_client.state.labware.get_lid_by_labware_id(
                labware.labware_id
            )
            if lid is not None:
                lid_id = lid.id
            else:
                raise ValueError("Cannot move a lid off of a labware with no lid.")

        _pick_up_offset = (
            LabwareOffsetVector(
                x=pick_up_offset[0], y=pick_up_offset[1], z=pick_up_offset[2]
            )
            if pick_up_offset
            else None
        )
        _drop_offset = (
            LabwareOffsetVector(x=drop_offset[0], y=drop_offset[1], z=drop_offset[2])
            if drop_offset
            else None
        )

        create_new_lid_stack = False

        if isinstance(new_location, DeckSlotName) or isinstance(
            new_location, StagingSlotName
        ):
            # Find the destination labware at the provided deck slot
            destination_labware_in_slot = self._engine_client.state.labware.get_by_slot(
                new_location
            )
            if destination_labware_in_slot is None:
                to_location = self._convert_labware_location(location=new_location)
                # absolutely must make a new lid stack
                create_new_lid_stack = True
            else:
                highest_child_location = (
                    self._engine_client.state.labware.get_highest_child_labware(
                        destination_labware_in_slot.id
                    )
                )
                if labware_validation.validate_definition_is_adapter(
                    self._engine_client.state.labware.get_definition(
                        highest_child_location
                    )
                ):
                    # absolutely must make a new lid stack
                    create_new_lid_stack = True

                to_location = self._convert_labware_location(
                    location=LabwareCore(highest_child_location, self._engine_client)
                )
        elif isinstance(new_location, LabwareCore):
            highest_child_location = (
                self._engine_client.state.labware.get_highest_child_labware(
                    new_location.labware_id
                )
            )
            if labware_validation.validate_definition_is_adapter(
                self._engine_client.state.labware.get_definition(highest_child_location)
            ):
                # absolutely must make a new lid stack
                create_new_lid_stack = True
            to_location = self._convert_labware_location(
                location=LabwareCore(highest_child_location, self._engine_client)
            )
        else:
            to_location = self._convert_labware_location(location=new_location)

        output_result = None
        if create_new_lid_stack:
            # Make a new lid stack object that is empty
            result = self._engine_client.execute_command_without_recovery(
                cmd.LoadLidStackParams(
                    location=SYSTEM_LOCATION,
                    loadName="empty",
                    version=1,
                    namespace="empty",
                    quantity=0,
                )
            )

            # Move the lid stack object from the SYSTEM_LOCATION space to the desired deck location
            self._engine_client.execute_command(
                cmd.MoveLabwareParams(
                    labwareId=result.stackLabwareId,
                    newLocation=to_location,
                    strategy=LabwareMovementStrategy.MANUAL_MOVE_WITHOUT_PAUSE,
                    pickUpOffset=None,
                    dropOffset=None,
                )
            )

            output_result = LabwareCore(
                labware_id=result.stackLabwareId, engine_client=self._engine_client
            )
            destination = self._convert_labware_location(location=output_result)
        else:
            destination = to_location

        self._engine_client.execute_command(
            cmd.MoveLabwareParams(
                labwareId=lid_id,
                newLocation=destination,
                strategy=strategy,
                pickUpOffset=_pick_up_offset,
                dropOffset=_drop_offset,
            )
        )

        # Handle leftover empty lid stack if there is one
        if (
            labware_validation.is_lid_stack(labware.load_name)
            and self._engine_client.state.labware.get_highest_child_labware(
                labware_id=labware.labware_id
            )
            == labware.labware_id
        ):
            # The originating lid stack is now empty, so we need to move it to the SYSTEM_LOCATION
            self._engine_client.execute_command(
                cmd.MoveLabwareParams(
                    labwareId=labware.labware_id,
                    newLocation=SYSTEM_LOCATION,
                    strategy=LabwareMovementStrategy.MANUAL_MOVE_WITHOUT_PAUSE,
                    pickUpOffset=None,
                    dropOffset=None,
                )
            )

        if strategy == LabwareMovementStrategy.USING_GRIPPER:
            # Clear out last location since it is not relevant to pipetting
            # and we only use last location for in-place pipetting commands
            self.set_last_location(location=None, mount=Mount.EXTENSION)

        # FIXME(jbl, 2024-01-04) deck conflict after execution logic issue, read notes in load_labware for more info:
        deck_conflict.check(
            engine_state=self._engine_client.state,
            new_labware_id=lid_id,
            existing_disposal_locations=self._disposal_locations,
            # TODO: We can now fetch these IDs from engine too.
            #  See comment in self.load_labware().
            existing_labware_ids=[
                labware_id
                for labware_id in self._labware_cores_by_id
                if labware_id != labware_id
            ],
            existing_module_ids=list(self._module_cores_by_id.keys()),
        )

        return output_result

    def _resolve_module_hardware(
        self, serial_number: str, model: ModuleModel
    ) -> AbstractModule:
        """Resolve a module serial number to module hardware API."""
        if self.is_simulating():
            return self._sync_hardware.create_simulating_module(model)  # type: ignore[no-any-return]

        for module_hardware in self._sync_hardware.attached_modules:
            if serial_number == module_hardware.device_info["serial"]:
                return module_hardware  # type: ignore[no-any-return]

        raise RuntimeError(f"Could not find specified module: {model.value}")

    def load_module(
        self,
        model: ModuleModel,
        deck_slot: Optional[DeckSlotName],
        configuration: Optional[str],
    ) -> Union[ModuleCore, NonConnectedModuleCore]:
        """Load a module into the protocol."""
        assert configuration is None, "Module `configuration` is deprecated"

        module_type = ModuleType.from_model(model)
        # TODO(mc, 2022-10-20): move to public ProtocolContext
        # once `Deck` and `ProtocolEngine` play nicely together
        if deck_slot is None:
            module_type = ModuleType.from_model(model)
            if module_type == ModuleType.THERMOCYCLER:
                deck_slot = DeckSlotName.SLOT_7
            else:
                raise InvalidModuleLocationError(deck_slot, model.name)

        robot_type = self._engine_client.state.config.robot_type
        # todo(mm, 2024-12-03): This might be possible to remove:
        # Protocol Engine will normalize the deck slot itself.
        normalized_deck_slot = deck_slot.to_equivalent_for_robot_type(robot_type)

        result = self._engine_client.execute_command_without_recovery(
            cmd.LoadModuleParams(
                model=EngineModuleModel(model),
                location=DeckSlotLocation(slotName=normalized_deck_slot),
            )
        )

        module_core = self._get_module_core(load_module_result=result, model=model)

        # FIXME(mm, 2023-02-21):
        # We're wrongly doing this conflict check *after* we've already loaded the
        # module into the ProtocolEngine. See FIXME comment in self.load_labware().
        deck_conflict.check(
            engine_state=self._engine_client.state,
            new_module_id=result.moduleId,
            existing_disposal_locations=self._disposal_locations,
            # TODO: We can now fetch these IDs from engine too.
            #  See comment in self.load_labware().
            #
            # Wrapping .keys() in list() is just to make Decoy verification easier.
            existing_labware_ids=list(self._labware_cores_by_id.keys()),
            existing_module_ids=list(self._module_cores_by_id.keys()),
        )

        self._module_cores_by_id[module_core.module_id] = module_core

        return module_core

    def _create_non_connected_module_core(
        self, load_module_result: LoadModuleResult
    ) -> NonConnectedModuleCore:
        return MagneticBlockCore(
            module_id=load_module_result.moduleId,
            engine_client=self._engine_client,
            api_version=self.api_version,
        )

    def _create_module_core(
        self, load_module_result: LoadModuleResult, model: ModuleModel
    ) -> ModuleCore:
        module_core_cls: Type[ModuleCore] = ModuleCore

        type_lookup: Dict[ModuleType, Type[ModuleCore]] = {
            ModuleType.TEMPERATURE: TemperatureModuleCore,
            ModuleType.MAGNETIC: MagneticModuleCore,
            ModuleType.THERMOCYCLER: ThermocyclerModuleCore,
            ModuleType.HEATER_SHAKER: HeaterShakerModuleCore,
            ModuleType.ABSORBANCE_READER: AbsorbanceReaderCore,
            ModuleType.FLEX_STACKER: FlexStackerCore,
        }

        module_type = load_module_result.model.as_type()

        module_core_cls = type_lookup[module_type]

        assert (
            load_module_result.serialNumber is not None
        ), "Expected a connected module but did not get a serial number."
        selected_hardware = self._resolve_module_hardware(
            load_module_result.serialNumber, model
        )

        return module_core_cls(
            module_id=load_module_result.moduleId,
            engine_client=self._engine_client,
            api_version=self.api_version,
            sync_module_hardware=SynchronousAdapter(selected_hardware),
        )

    def _get_module_core(
        self, load_module_result: LoadModuleResult, model: ModuleModel
    ) -> Union[ModuleCore, NonConnectedModuleCore]:
        if ProtocolEngineModuleModel.is_magnetic_block(load_module_result.model):
            return self._create_non_connected_module_core(load_module_result)
        else:
            return self._create_module_core(
                load_module_result=load_module_result, model=model
            )

    def add_or_get_labware_core(self, labware_id: str) -> LabwareCore:
        """Create a LabwareCore and add it to the map or return one if it exists."""
        if labware_id in self._labware_cores_by_id:
            return self._labware_cores_by_id[labware_id]
        else:
            core = LabwareCore(labware_id, self._engine_client)
            self._labware_cores_by_id[labware_id] = core
            return core

    def load_robot(self) -> RobotCore:
        """Load a robot core into the RobotContext."""
        return RobotCore(
            engine_client=self._engine_client, sync_hardware_api=self._sync_hardware
        )

    def load_instrument(
        self,
        instrument_name: PipetteNameType,
        mount: Mount,
        liquid_presence_detection: bool = False,
    ) -> InstrumentCore:
        """Load an instrument into the protocol.

        Args:
            instrument_name: Load name of the instrument.
            mount: Mount the instrument is attached to.

        Returns:
            An instrument core configured to use the requested instrument.
        """
        engine_mount = MountType[mount.name]
        load_result = self._engine_client.execute_command_without_recovery(
            cmd.LoadPipetteParams(
                pipetteName=instrument_name,
                mount=engine_mount,
                tipOverlapNotAfterVersion=overlap_versions.overlap_for_api_version(
                    self._api_version
                ),
                liquidPresenceDetection=liquid_presence_detection,
            )
        )

        return InstrumentCore(
            pipette_id=load_result.pipetteId,
            engine_client=self._engine_client,
            sync_hardware_api=self._sync_hardware,
            protocol_core=self,
            # TODO(mm, 2022-11-10): Deduplicate "400" with legacy core.
            default_movement_speed=400,
        )

    def load_trash_bin(self, slot_name: DeckSlotName, area_name: str) -> TrashBin:
        """Load a deck configuration based trash bin.

        Args:
            slot_name: the slot the trash is being loaded into.
            area_name: the addressable area name of the trash.

        Returns:
            A trash bin object.
        """
        trash_bin = TrashBin(
            location=slot_name,
            addressable_area_name=area_name,
            api_version=self._api_version,
            engine_client=self._engine_client,
        )
        self._add_disposal_location_to_engine(trash_bin)
        return trash_bin

    def load_ot2_fixed_trash_bin(self) -> None:
        """Load a deck configured OT-2 fixed trash in Slot 12."""
        _fixed_trash_trash_bin = TrashBin(
            location=DeckSlotName.FIXED_TRASH,
            addressable_area_name="fixedTrash",
            api_version=self._api_version,
            engine_client=self._engine_client,
        )
        # We are just appending the fixed trash to the core's internal list here, not adding it to the engine via
        # the core, since that method works through the SyncClient and if called from here, will cause protocols
        # to deadlock. Instead, that method is called in protocol engine directly in create_protocol_context after
        # ProtocolContext is initialized.
        self.append_disposal_location(_fixed_trash_trash_bin)

    def load_waste_chute(self) -> WasteChute:
        """Load a deck configured waste chute into Slot D3.

        Returns:
            A waste chute object.
        """
        waste_chute = WasteChute(
            engine_client=self._engine_client, api_version=self._api_version
        )
        self._add_disposal_location_to_engine(waste_chute)
        return waste_chute

    def pause(self, msg: Optional[str]) -> None:
        """Pause the protocol."""
        self._engine_client.execute_command(cmd.WaitForResumeParams(message=msg))

    def comment(self, msg: str) -> None:
        """Create a comment in the protocol to be shown in the log."""
        self._engine_client.execute_command(cmd.CommentParams(message=msg))

    def delay(self, seconds: float, msg: Optional[str]) -> None:
        """Wait for a period of time before proceeding."""
        self._engine_client.execute_command(
            cmd.WaitForDurationParams(seconds=seconds, message=msg)
        )

    def home(self) -> None:
        """Move all axes to their home positions."""
        self._engine_client.execute_command(cmd.HomeParams(axes=None))

    def set_rail_lights(self, on: bool) -> None:
        """Set the device's rail lights."""
        self._engine_client.execute_command(cmd.SetRailLightsParams(on=on))

    def get_rail_lights_on(self) -> bool:
        """Get whether the device's rail lights are on."""
        return self._sync_hardware.get_lights()["rails"]  # type: ignore[no-any-return]

    def door_closed(self) -> bool:
        """Get whether the device's front door is closed."""
        return self._sync_hardware.door_state == DoorState.CLOSED  # type: ignore[no-any-return]

    def get_last_location(
        self,
        mount: Optional[Mount] = None,
    ) -> Optional[Union[Location, TrashBin, WasteChute]]:
        """Get the last accessed location."""
        if mount is None or mount == self._last_mount:
            return self._last_location

        return None

    def set_last_location(
        self,
        location: Optional[Union[Location, TrashBin, WasteChute]],
        mount: Optional[Mount] = None,
    ) -> None:
        """Set the last accessed location."""
        self._last_location = location
        self._last_mount = mount

    def load_lid_stack(
        self,
        load_name: str,
        location: Union[DeckSlotName, StagingSlotName, LabwareCore],
        quantity: int,
        namespace: Optional[str],
        version: Optional[int],
    ) -> LabwareCore:
        """Load a Stack of Lids to a given location, creating a Lid Stack."""
        if quantity < 1:
            raise ValueError(
                "When loading a lid stack quantity cannot be less than one."
            )
        if isinstance(location, DeckSlotName) or isinstance(location, StagingSlotName):
            load_location = self._convert_labware_location(location=location)
        else:
            if isinstance(location, LabwareCore):
                load_location = self._convert_labware_location(location=location)
            else:
                raise ValueError(
                    "Expected type of Labware Location for lid stack must be Labware, not Legacy Labware or Well."
                )

        custom_labware_params = (
            self._engine_client.state.labware.find_custom_labware_load_params()
        )
        namespace, version = load_labware_params.resolve(
            load_name, namespace, version, custom_labware_params, self._api_version
        )

        load_result = self._engine_client.execute_command_without_recovery(
            cmd.LoadLidStackParams(
                loadName=load_name,
                location=load_location,
                namespace=namespace,
                version=version,
                quantity=quantity,
            )
        )

        # FIXME(CHB, 2024-12-04) just like load labware and load adapter we have a validating after loading the object issue
        assert load_result.definition is not None
        validation.ensure_definition_is_lid(load_result.definition)

        deck_conflict.check(
            engine_state=self._engine_client.state,
            new_labware_id=load_result.stackLabwareId,
            existing_disposal_locations=self._disposal_locations,
            # TODO (spp, 2023-11-27): We've been using IDs from _labware_cores_by_id
            #  and _module_cores_by_id instead of getting the lists directly from engine
            #  because of the chance of engine carrying labware IDs from LPC too.
            #  But with https://github.com/Opentrons/opentrons/pull/13943,
            #  & LPC in maintenance runs, we can now rely on engine state for these IDs too.
            # Wrapping .keys() in list() is just to make Decoy verification easier.
            existing_labware_ids=list(self._labware_cores_by_id.keys()),
            existing_module_ids=list(self._module_cores_by_id.keys()),
        )

        labware_core = LabwareCore(
            labware_id=load_result.stackLabwareId,
            engine_client=self._engine_client,
        )

        self._labware_cores_by_id[labware_core.labware_id] = labware_core

        return labware_core

    def get_deck_definition(self) -> DeckDefinitionV5:
        """Get the geometry definition of the robot's deck."""
        return self._engine_client.state.labware.get_deck_definition()

    def get_slot_definition(
        self, slot: Union[DeckSlotName, StagingSlotName]
    ) -> SlotDefV3:
        """Get the slot definition from the robot's deck."""
        return self._engine_client.state.addressable_areas.get_slot_definition(slot.id)

    def get_slot_definitions(self) -> Dict[str, SlotDefV3]:
        """Get all standard slot definitions available in the deck definition."""
        return self._engine_client.state.addressable_areas.get_deck_slot_definitions()

    def get_staging_slot_definitions(self) -> Dict[str, SlotDefV3]:
        """Get all staging slot definitions available in the deck definition."""
        return (
            self._engine_client.state.addressable_areas.get_staging_slot_definitions()
        )

    def get_slot_item(
        self, slot_name: Union[DeckSlotName, StagingSlotName]
    ) -> Union[LabwareCore, ModuleCore, NonConnectedModuleCore, None]:
        """Get the contents of a given slot, if any."""
        loaded_item = self._engine_client.state.geometry.get_slot_item(
            slot_name=slot_name
        )

        if isinstance(loaded_item, LoadedLabware):
            return self._labware_cores_by_id[loaded_item.id]

        if isinstance(loaded_item, LoadedModule):
            return self._module_cores_by_id[loaded_item.id]

        return None

    def get_labware_on_module(
        self, module_core: Union[ModuleCore, NonConnectedModuleCore]
    ) -> Optional[LabwareCore]:
        """Get the item on top of a given module, if any."""
        try:
            labware_id = self._engine_client.state.labware.get_id_by_module(
                module_core.module_id
            )
        except LabwareNotLoadedOnModuleError:
            return None
        return self.add_or_get_labware_core(labware_id)

    def get_labware_on_labware(
        self, labware_core: LabwareCore
    ) -> Optional[LabwareCore]:
        """Get the item on top of a given labware, if any."""
        try:
            labware_id = self._engine_client.state.labware.get_id_by_labware(
                labware_core.labware_id
            )
        except LabwareNotLoadedOnLabwareError:
            return None
        return self.add_or_get_labware_core(labware_id)

    def get_slot_center(self, slot_name: Union[DeckSlotName, StagingSlotName]) -> Point:
        """Get the absolute coordinate of a slot's center."""
        return self._engine_client.state.addressable_areas.get_addressable_area_center(
            slot_name.id
        )

    def get_highest_z(self) -> float:
        """Get the highest Z point of all deck items."""
        return self._engine_client.state.geometry.get_all_obstacle_highest_z()

    def get_labware_cores(self) -> List[LabwareCore]:
        """Get all loaded labware cores."""
        return list(self._labware_cores_by_id.values())

    def get_module_cores(self) -> List[Union[ModuleCore, NonConnectedModuleCore]]:
        """Get all loaded module cores."""
        return list(self._module_cores_by_id.values())

    def define_liquid(
        self,
        name: str,
        description: Optional[str],
        display_color: Optional[str],
    ) -> Liquid:
        """Define a liquid to load into a well."""
        liquid = self._engine_client.add_liquid(
            name=name, description=description, color=display_color
        )

        return Liquid(
            _id=liquid.id,
            name=liquid.displayName,
            description=liquid.description,
            display_color=(liquid.displayColor.root if liquid.displayColor else None),
        )

    def get_liquid_class(self, name: str, version: int) -> LiquidClass:
        """Get an instance of a built-in liquid class."""
        try:
            # Check if we have already loaded this liquid class' definition
            liquid_class_def = self._liquid_class_def_cache[(name, version)]
        except KeyError:
            try:
                # Fetching the liquid class data from file and parsing it
                # is an expensive operation and should be avoided.
                # Calling this often will degrade protocol execution performance.
                liquid_class_def = liquid_classes.load_definition(name, version=version)
                self._liquid_class_def_cache[(name, version)] = liquid_class_def
            except LiquidClassDefinitionDoesNotExist:
                raise ValueError(
                    f"Liquid class definition not found for '{name}' version {version}."
                )

        return LiquidClass.create(liquid_class_def)

    def get_labware_location(
        self, labware_core: LabwareCore
    ) -> Union[str, LabwareCore, ModuleCore, NonConnectedModuleCore, OffDeckType]:
        """Get labware parent location."""
        labware_location = self._engine_client.state.labware.get_location(
            labware_core.labware_id
        )
        if isinstance(labware_location, DeckSlotLocation):
            return validation.internal_slot_to_public_string(
                labware_location.slotName, self._engine_client.state.config.robot_type
            )
        elif isinstance(labware_location, AddressableAreaLocation):
            # This will only ever be a robot accurate deck slot name or Flex staging slot name
            return labware_location.addressableAreaName
        elif isinstance(labware_location, ModuleLocation):
            return self._module_cores_by_id[labware_location.moduleId]
        elif isinstance(labware_location, OnLabwareLocation):
            return self._labware_cores_by_id[labware_location.labwareId]

        return OffDeckType.OFF_DECK

    def _convert_labware_location(
        self,
        location: Union[
            DeckSlotName,
            StagingSlotName,
            LabwareCore,
            ModuleCore,
            NonConnectedModuleCore,
            OffDeckType,
            WasteChute,
            TrashBin,
        ],
    ) -> LoadableLabwareLocation:
        if isinstance(location, LabwareCore):
            return OnLabwareLocation(labwareId=location.labware_id)
        else:
            return self._get_non_stacked_location(location)

    @staticmethod
    def _get_non_stacked_location(
        location: Union[
            DeckSlotName,
            StagingSlotName,
            ModuleCore,
            NonConnectedModuleCore,
            OffDeckType,
            WasteChute,
            TrashBin,
        ],
    ) -> NonStackedLocation:
        if isinstance(location, (ModuleCore, NonConnectedModuleCore)):
            return ModuleLocation(moduleId=location.module_id)
        elif location is OffDeckType.OFF_DECK:
            return OFF_DECK_LOCATION
        elif isinstance(location, DeckSlotName):
            return DeckSlotLocation(slotName=location)
        elif isinstance(location, StagingSlotName):
            return AddressableAreaLocation(addressableAreaName=location.id)
        elif isinstance(location, WasteChute):
            # TODO(mm, 2023-12-06) This will need to determine the appropriate Waste Chute to return, but only move_labware uses this for now
            return AddressableAreaLocation(addressableAreaName="gripperWasteChute")
        elif isinstance(location, TrashBin):
            return AddressableAreaLocation(addressableAreaName=location.area_name)



from typing import List, Optional, Protocol

from .types import (
    ActiveRange,
    LEDPattern,
    LimitSwitchStatus,
    MeasurementKind,
    MoveResult,
    SpadMapID,
    StackerAxis,
    PlatformStatus,
    Direction,
    MoveParams,
    StackerInfo,
    LEDColor,
    StallGuardParams,
    TOFConfiguration,
    TOFMeasurement,
    TOFMeasurementResult,
    TOFSensor,
    TOFSensorStatus,
)


class AbstractFlexStackerDriver(Protocol):
    """Protocol for the Stacker driver."""

    async def connect(self) -> None:
        """Connect to stacker."""
        ...

    async def disconnect(self) -> None:
        """Disconnect from stacker."""
        ...

    async def is_connected(self) -> bool:
        """Check connection to stacker."""
        ...

    async def get_device_info(self) -> StackerInfo:
        """Get Device Info."""
        ...

    async def set_serial_number(self, sn: str) -> None:
        """Set Serial Number."""
        ...

    async def enable_motors(self, axis: List[StackerAxis]) -> None:
        """Enables the axis motor if present, disables it otherwise."""
        ...

    async def stop_motors(self) -> None:
        """Stop all motor movement."""
        ...

    async def set_run_current(self, axis: StackerAxis, current: float) -> None:
        """Set axis peak run current in amps."""
        ...

    async def set_ihold_current(self, axis: StackerAxis, current: float) -> None:
        """Set axis hold current in amps."""
        ...

    async def set_stallguard_threshold(
        self, axis: StackerAxis, enable: bool, threshold: int
    ) -> None:
        """Enables and sets the stallguard threshold for the given axis motor."""
        ...

    async def enable_tof_sensor(self, sensor: TOFSensor, enable: bool) -> None:
        """Enable or disable the TOF sensor."""
        ...

    async def manage_tof_measurement(
        self,
        sensor: TOFSensor,
        kind: MeasurementKind = MeasurementKind.HISTOGRAM,
        start: bool = True,
    ) -> TOFMeasurement:
        """Start or stop Measurements from the TOF sensor."""
        ...

    async def get_tof_histogram(self, sensor: TOFSensor) -> TOFMeasurementResult:
        """Get the full histogram measurement from the TOF sensor."""
        ...

    async def set_tof_configuration(
        self,
        sensor: TOFSensor,
        spad_map_id: SpadMapID,
        active_range: Optional[ActiveRange] = None,
        kilo_iterations: Optional[int] = None,
        report_period_ms: Optional[int] = None,
        histogram_dump: Optional[bool] = None,
    ) -> None:
        """Set the configuration of the TOF sensor.

        :param sensor: The TOF sensor to configure.
        :param spad_map_id: The pre-defined SPAD map which sets the fov and focus area (14 default).
        :active_range: The operating mode Short-range high-accuracy (default) or long range.
        :kilo_iterations: The Measurement iterations times 1024 (4000 default).
        :report_period_ms: The reporting period before each measurement (500 default).
        :histogram_dump: Enables/Disables histogram measurements (True default).
        :return: None
        """
        ...

    async def get_tof_configuration(self, sensor: TOFSensor) -> TOFConfiguration:
        """Get the configuration of the TOF sensor."""
        ...

    async def set_motor_driver_register(
        self, axis: StackerAxis, reg: int, value: int
    ) -> None:
        """Set the register of the given motor axis driver to the given value."""
        ...

    async def get_motor_driver_register(self, axis: StackerAxis, reg: int) -> int:
        """Gets the register value of the given motor axis driver."""
        ...

    async def set_tof_driver_register(
        self, sensor: TOFSensor, reg: int, value: int
    ) -> None:
        """Set the register of the given tof sensor driver to the given value."""
        ...

    async def get_tof_driver_register(self, sensor: TOFSensor, reg: int) -> int:
        """Gets the register value of the given tof sensor driver."""
        ...

    async def get_tof_sensor_status(self, sensor: TOFSensor) -> TOFSensorStatus:
        """Get the status of the tof sensor."""
        ...

    async def get_motion_params(self, axis: StackerAxis) -> MoveParams:
        """Get the motion parameters used by the given axis motor."""
        ...

    async def get_stallguard_threshold(self, axis: StackerAxis) -> StallGuardParams:
        """Get the stallguard parameters by the given axis motor."""
        ...

    async def get_limit_switch(self, axis: StackerAxis, direction: Direction) -> bool:
        """Get limit switch status.

        :return: True if limit switch is triggered, False otherwise
        """
        ...

    async def get_limit_switches_status(self) -> LimitSwitchStatus:
        """Get limit switch statuses for all axes."""
        ...

    async def get_platform_sensor(self, direction: Direction) -> bool:
        """Get platform sensor status.

        :return: True if platform is present, False otherwise
        """
        ...

    async def get_platform_status(self) -> PlatformStatus:
        """Get platform status."""
        ...

    async def get_hopper_door_closed(self) -> bool:
        """Get whether or not door is closed.

        :return: True if door is closed, False otherwise
        """
        ...

    async def get_installation_detected(self) -> bool:
        """Get whether or not installation is detected.

        :return: True if installation is detected, False otherwise
        """
        ...

    async def move_in_mm(
        self, axis: StackerAxis, distance: float, params: MoveParams | None = None
    ) -> MoveResult:
        """Move axis by the given distance in mm."""
        ...

    async def move_to_limit_switch(
        self, axis: StackerAxis, direction: Direction, params: MoveParams | None = None
    ) -> MoveResult:
        """Move until limit switch is triggered."""
        ...

    async def home_axis(self, axis: StackerAxis, direction: Direction) -> MoveResult:
        """Home axis."""
        ...

    async def set_led(
        self,
        power: float,
        color: Optional[LEDColor] = None,
        external: Optional[bool] = None,
        pattern: Optional[LEDPattern] = None,
        duration: Optional[int] = None,
        reps: Optional[int] = None,
    ) -> None:
        """Set LED Status bar color and pattern."""
        ...

    async def enter_programming_mode(self) -> None:
        """Reboot into programming mode"""
        ...

    def reset_serial_buffers(self) -> None:
        """Reset the input and output serial buffers."""
        ...



"""
USB Simulating Driver.

A class to convert info from the usb bus into a
more readable format.
"""
from typing import List, Union

from opentrons.hardware_control.modules.types import (
    ModuleAtPort,
    SimulatingModuleAtPort,
)

from .interfaces import USBDriverInterface


class USBBusSimulator(USBDriverInterface):
    def match_virtual_ports(
        self,
        virtual_port: Union[List[ModuleAtPort], List[SimulatingModuleAtPort]],
    ) -> Union[List[ModuleAtPort], List[SimulatingModuleAtPort]]:
        return virtual_port



"""Calibrate-module command for OT3 hardware. request, result, and implementation models."""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal
from pydantic import BaseModel, Field

from opentrons.types import MountType
from opentrons.protocol_engine.resources.ot3_validation import ensure_ot3_hardware
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence

# Work around type-only circular dependencies.
if TYPE_CHECKING:
    from ...state.state import StateView

from ...types import ModuleOffsetVector, DeckSlotLocation

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import OT3Mount
from opentrons.hardware_control import ot3_calibration as calibration


CalibrateModuleCommandType = Literal["calibration/calibrateModule"]


class CalibrateModuleParams(BaseModel):
    """Payload required to calibrate-module."""

    moduleId: str = Field(..., description="The unique id of module to calibrate.")
    labwareId: str = Field(
        ..., description="The unique id of module calibration adapter labware."
    )
    mount: MountType = Field(
        ..., description="The instrument mount used to calibrate the module."
    )


class CalibrateModuleResult(BaseModel):
    """Result data from the execution of a calibrate-module command."""

    moduleOffset: ModuleOffsetVector = Field(
        ..., description="Offset of calibrated module."
    )

    location: DeckSlotLocation = Field(
        ..., description="The deck slot this module was calibrated in."
    )


class CalibrateModuleImplementation(
    AbstractCommandImpl[CalibrateModuleParams, SuccessData[CalibrateModuleResult]]
):
    """CalibrateModule command implementation."""

    def __init__(
        self,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._hardware_api = hardware_api

    async def execute(
        self, params: CalibrateModuleParams
    ) -> SuccessData[CalibrateModuleResult]:
        """Execute calibrate-module command."""
        ot3_api = ensure_ot3_hardware(
            self._hardware_api,
        )
        ot3_mount = OT3Mount.from_mount(params.mount)
        slot = self._state_view.modules.get_location(params.moduleId)
        module_serial = self._state_view.modules.get_serial_number(params.moduleId)
        # NOTE (ba, 2023-03-31): There are two wells for calibration labware definitions
        # well A1 represents the location calibration square center relative to the adapters bottom-left corner
        # well B1 represents the location of the calibration square probe point relative to the adapters bottom-left corner.
        nominal_position = self._state_view.geometry.get_nominal_well_position(
            labware_id=params.labwareId, well_name="B1"
        )

        # start the calibration
        module_offset = await calibration.calibrate_module(
            ot3_api, ot3_mount, slot.slotName.id, module_serial, nominal_position
        )

        return SuccessData(
            public=CalibrateModuleResult(
                moduleOffset=ModuleOffsetVector(
                    x=module_offset.x, y=module_offset.y, z=module_offset.z
                ),
                location=slot,
            ),
        )


class CalibrateModule(
    BaseCommand[CalibrateModuleParams, CalibrateModuleResult, ErrorOccurrence]
):
    """Calibrate-module command model."""

    commandType: CalibrateModuleCommandType = "calibration/calibrateModule"
    params: CalibrateModuleParams
    result: Optional[CalibrateModuleResult] = None

    _ImplementationCls: Type[
        CalibrateModuleImplementation
    ] = CalibrateModuleImplementation


class CalibrateModuleCreate(BaseCommandCreate[CalibrateModuleParams]):
    """Create calibrate-module command request model."""

    commandType: CalibrateModuleCommandType = "calibration/calibrateModule"
    params: CalibrateModuleParams

    _CommandCls: Type[CalibrateModule] = CalibrateModule



"""The interface that implements ProtocolContext."""

from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Generic, List, Optional, Union, Tuple, Dict, TYPE_CHECKING

from opentrons_shared_data.deck.types import DeckDefinitionV5, SlotDefV3
from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons_shared_data.labware.types import LabwareDefinition
from opentrons_shared_data.robot.types import RobotType

from opentrons.types import (
    DeckSlotName,
    StagingSlotName,
    Location,
    Mount,
    Point,
)
from opentrons.hardware_control import SyncHardwareAPI
from opentrons.hardware_control.modules.types import ModuleModel
from opentrons.protocols.api_support.util import AxisMaxSpeeds

from .instrument import InstrumentCoreType
from .labware import LabwareCoreType, LabwareLoadParams
from .module import ModuleCoreType
from .._liquid import Liquid, LiquidClass
from .robot import AbstractRobot
from .._types import OffDeckType
from ..disposal_locations import TrashBin, WasteChute

if TYPE_CHECKING:
    from ..labware import Labware


class AbstractProtocol(
    ABC, Generic[InstrumentCoreType, LabwareCoreType, ModuleCoreType]
):
    @property
    @abstractmethod
    def fixed_trash(self) -> Optional[LabwareCoreType]:
        """Get the fixed trash labware core."""
        ...

    @property
    @abstractmethod
    def robot_type(self) -> RobotType:
        ...

    @abstractmethod
    def get_max_speeds(self) -> AxisMaxSpeeds:
        ...

    @abstractmethod
    def get_hardware(self) -> SyncHardwareAPI:
        ...

    @abstractmethod
    def is_simulating(self) -> bool:
        ...

    @abstractmethod
    def add_labware_definition(
        self,
        definition: LabwareDefinition,
    ) -> LabwareLoadParams:
        """Add a labware definition to the set of loadable definitions."""
        ...

    @abstractmethod
    def append_disposal_location(
        self,
        disposal_location: Union[Labware, TrashBin, WasteChute],
    ) -> None:
        """Append a disposal location object to the core"""
        ...

    @abstractmethod
    def load_labware(
        self,
        load_name: str,
        location: Union[
            DeckSlotName, StagingSlotName, LabwareCoreType, ModuleCoreType, OffDeckType
        ],
        label: Optional[str],
        namespace: Optional[str],
        version: Optional[int],
    ) -> LabwareCoreType:
        """Load a labware using its identifying parameters."""
        ...

    @abstractmethod
    def load_adapter(
        self,
        load_name: str,
        location: Union[DeckSlotName, StagingSlotName, ModuleCoreType, OffDeckType],
        namespace: Optional[str],
        version: Optional[int],
    ) -> LabwareCoreType:
        """Load an adapter using its identifying parameters"""
        ...

    @abstractmethod
    def load_lid(
        self,
        load_name: str,
        location: LabwareCoreType,
        namespace: Optional[str],
        version: Optional[int],
    ) -> LabwareCoreType:
        """Load an individual lid labware using its identifying parameters. Must be loaded on a labware."""
        ...

    @abstractmethod
    def load_labware_to_flex_stacker_hopper(
        self,
        module_core: ModuleCoreType,
        load_name: str,
        quantity: int,
        label: Optional[str],
        namespace: Optional[str],
        version: Optional[int],
        lid: Optional[str],
    ) -> None:
        """Load one or more labware with or without a lid to the flex stacker hopper."""
        ...

    @abstractmethod
    def move_labware(
        self,
        labware_core: LabwareCoreType,
        new_location: Union[
            DeckSlotName,
            StagingSlotName,
            LabwareCoreType,
            ModuleCoreType,
            OffDeckType,
            WasteChute,
            TrashBin,
        ],
        use_gripper: bool,
        pause_for_manual_move: bool,
        pick_up_offset: Optional[Tuple[float, float, float]],
        drop_offset: Optional[Tuple[float, float, float]],
    ) -> None:
        ...

    @abstractmethod
    def move_lid(
        self,
        source_location: Union[DeckSlotName, StagingSlotName, LabwareCoreType],
        new_location: Union[
            DeckSlotName,
            StagingSlotName,
            LabwareCoreType,
            OffDeckType,
            WasteChute,
            TrashBin,
        ],
        use_gripper: bool,
        pause_for_manual_move: bool,
        pick_up_offset: Optional[Tuple[float, float, float]],
        drop_offset: Optional[Tuple[float, float, float]],
    ) -> LabwareCoreType | None:
        ...

    @abstractmethod
    def load_module(
        self,
        model: ModuleModel,
        deck_slot: Optional[DeckSlotName],
        configuration: Optional[str],
    ) -> ModuleCoreType:
        ...

    @abstractmethod
    def load_instrument(
        self,
        instrument_name: PipetteNameType,
        mount: Mount,
        liquid_presence_detection: bool = False,
    ) -> InstrumentCoreType:
        ...

    @abstractmethod
    def load_trash_bin(self, slot_name: DeckSlotName, area_name: str) -> TrashBin:
        ...

    @abstractmethod
    def load_ot2_fixed_trash_bin(self) -> None:
        ...

    @abstractmethod
    def load_waste_chute(self) -> WasteChute:
        ...

    @abstractmethod
    def pause(self, msg: Optional[str]) -> None:
        ...

    @abstractmethod
    def comment(self, msg: str) -> None:
        ...

    @abstractmethod
    def delay(self, seconds: float, msg: Optional[str]) -> None:
        ...

    @abstractmethod
    def home(self) -> None:
        ...

    @abstractmethod
    def set_rail_lights(self, on: bool) -> None:
        ...

    @abstractmethod
    def get_disposal_locations(self) -> List[Union[Labware, TrashBin, WasteChute]]:
        ...

    @abstractmethod
    def get_rail_lights_on(self) -> bool:
        ...

    @abstractmethod
    def door_closed(self) -> bool:
        ...

    @abstractmethod
    def get_last_location(
        self,
        mount: Optional[Mount] = None,
    ) -> Optional[Union[Location, TrashBin, WasteChute]]:
        ...

    @abstractmethod
    def set_last_location(
        self,
        location: Optional[Location],
        mount: Optional[Mount] = None,
    ) -> None:
        ...

    @abstractmethod
    def load_lid_stack(
        self,
        load_name: str,
        location: Union[DeckSlotName, StagingSlotName, LabwareCoreType],
        quantity: int,
        namespace: Optional[str],
        version: Optional[int],
    ) -> LabwareCoreType:
        ...

    @abstractmethod
    def get_deck_definition(self) -> DeckDefinitionV5:
        """Get the geometry definition of the robot's deck."""

    @abstractmethod
    def get_slot_definition(
        self, slot: Union[DeckSlotName, StagingSlotName]
    ) -> SlotDefV3:
        """Get the slot definition from the robot's deck."""

    @abstractmethod
    def get_slot_definitions(self) -> Dict[str, SlotDefV3]:
        """Get all standard slot definitions available in the deck definition."""

    @abstractmethod
    def get_staging_slot_definitions(self) -> Dict[str, SlotDefV3]:
        """Get all staging slot definitions available in the deck definition."""

    @abstractmethod
    def get_slot_item(
        self, slot_name: Union[DeckSlotName, StagingSlotName]
    ) -> Union[LabwareCoreType, ModuleCoreType, None]:
        """Get the contents of a given slot, if any."""

    @abstractmethod
    def get_labware_on_module(
        self, module_core: ModuleCoreType
    ) -> Optional[LabwareCoreType]:
        """Get the labware on a given module, if any."""

    @abstractmethod
    def get_labware_on_labware(
        self, labware_core: LabwareCoreType
    ) -> Optional[LabwareCoreType]:
        """Get the labware on a given labware, if any."""

    @abstractmethod
    def get_slot_center(self, slot_name: Union[DeckSlotName, StagingSlotName]) -> Point:
        """Get the absolute coordinate of a slot's center."""

    @abstractmethod
    def get_highest_z(self) -> float:
        """Get the highest Z point of all deck items."""

    @abstractmethod
    def get_labware_cores(self) -> List[LabwareCoreType]:
        """Get all loaded labware cores."""

    @abstractmethod
    def get_module_cores(self) -> List[ModuleCoreType]:
        """Get all loaded module cores."""

    @abstractmethod
    def define_liquid(
        self, name: str, description: Optional[str], display_color: Optional[str]
    ) -> Liquid:
        """Define a liquid to load into a well."""

    @abstractmethod
    def get_liquid_class(self, name: str, version: int) -> LiquidClass:
        """Get an instance of a built-in liquid class."""

    @abstractmethod
    def get_labware_location(
        self, labware_core: LabwareCoreType
    ) -> Union[str, LabwareCoreType, ModuleCoreType, OffDeckType]:
        """Get labware parent location."""

    @abstractmethod
    def load_robot(self) -> AbstractRobot:
        """Load a Robot Core context into a protocol"""



"""Move to well command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .pipetting_common import (
    PipetteIdMixin,
)
from .movement_common import (
    LiquidHandlingWellLocationMixin,
    MovementMixin,
    DestinationPositionResult,
    StallOrCollisionError,
    move_to_well,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)

if TYPE_CHECKING:
    from ..execution import MovementHandler
    from ..state.state import StateView
    from ..resources.model_utils import ModelUtils

MoveToWellCommandType = Literal["moveToWell"]


class MoveToWellParams(PipetteIdMixin, LiquidHandlingWellLocationMixin, MovementMixin):
    """Payload required to move a pipette to a specific well."""

    pass


class MoveToWellResult(DestinationPositionResult):
    """Result data from the execution of a MoveToWell command."""

    pass


class MoveToWellImplementation(
    AbstractCommandImpl[
        MoveToWellParams,
        SuccessData[MoveToWellResult] | DefinedErrorData[StallOrCollisionError],
    ]
):
    """Move to well command implementation."""

    def __init__(
        self,
        state_view: StateView,
        movement: MovementHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._movement = movement
        self._model_utils = model_utils

    async def execute(
        self, params: MoveToWellParams
    ) -> SuccessData[MoveToWellResult] | DefinedErrorData[StallOrCollisionError]:
        """Move the requested pipette to the requested well."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName
        well_location = params.wellLocation
        # TODO(cm): implement move_to_well with meniscus + volume offset
        if well_location.volumeOffset and well_location.volumeOffset != 0:
            raise ValueError("volume offset not supported with MoveToWell")

        move_result = await move_to_well(
            model_utils=self._model_utils,
            movement=self._movement,
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            force_direct=params.forceDirect,
            minimum_z_height=params.minimumZHeight,
            speed=params.speed,
        )
        if isinstance(move_result, DefinedErrorData):
            return move_result
        else:
            return SuccessData(
                public=MoveToWellResult(position=move_result.public.position),
                state_update=move_result.state_update,
            )


class MoveToWell(
    BaseCommand[MoveToWellParams, MoveToWellResult, StallOrCollisionError]
):
    """Move to well command model."""

    commandType: MoveToWellCommandType = "moveToWell"
    params: MoveToWellParams
    result: Optional[MoveToWellResult] = None

    _ImplementationCls: Type[MoveToWellImplementation] = MoveToWellImplementation


class MoveToWellCreate(BaseCommandCreate[MoveToWellParams]):
    """Move to well command creation request model."""

    commandType: MoveToWellCommandType = "moveToWell"
    params: MoveToWellParams

    _CommandCls: Type[MoveToWell] = MoveToWell



"""Thermocycler Module sub-state."""
from dataclasses import dataclass
from typing import NewType, Optional

from opentrons.protocol_engine.errors import (
    InvalidTargetTemperatureError,
    InvalidBlockVolumeError,
    NoTargetTemperatureSetError,
    InvalidHoldTimeError,
)

# TODO(mc, 2022-04-25): move to module definition
# https://github.com/Opentrons/opentrons/issues/9800
from opentrons.drivers.thermocycler.driver import (
    BLOCK_TARGET_MIN,
    BLOCK_TARGET_MAX,
    BLOCK_VOL_MIN,
    BLOCK_VOL_MAX,
    LID_TARGET_MIN,
    LID_TARGET_MAX,
)
from opentrons.hardware_control.modules import ModuleData, ModuleDataValidator

ThermocyclerModuleId = NewType("ThermocyclerModuleId", str)


@dataclass(frozen=True)
class ThermocyclerModuleSubState:
    """Thermocycler-specific state.

    Provides calculations and read-only state access
    for an individual loaded Thermocycler Module.
    """

    module_id: ThermocyclerModuleId
    is_lid_open: bool
    target_block_temperature: Optional[float]
    target_lid_temperature: Optional[float]

    @staticmethod
    def validate_target_block_temperature(celsius: float) -> float:
        """Validate a given target block temperature.

        Args:
            celsius: The requested block temperature.

        Raises:
            InvalidTargetTemperatureError: The given temperature
                is outside the thermocycler's operating range.

        Returns:
            The validated temperature in degrees Celsius.
        """
        if BLOCK_TARGET_MIN <= celsius <= BLOCK_TARGET_MAX:
            return celsius

        raise InvalidTargetTemperatureError(
            "Thermocycler block temperature must be between"
            f" {BLOCK_TARGET_MIN} and {BLOCK_TARGET_MAX}, but got {celsius}."
        )

    @staticmethod
    def validate_max_block_volume(volume: float) -> float:
        """Validate a given target block max volume.

        Args:
            volume: The requested block max volume in uL.

        Raises:
            InvalidBlockVolumeError: The given volume
                is outside the thermocycler's operating range.

        Returns:
            The validated volume in uL.
        """
        if BLOCK_VOL_MIN <= volume <= BLOCK_VOL_MAX:
            return volume

        raise InvalidBlockVolumeError(
            "Thermocycler max block volume must be between"
            f" {BLOCK_VOL_MIN} and {BLOCK_VOL_MAX}, but got {volume}."
        )

    @staticmethod
    def validate_hold_time(hold_time: float) -> float:
        """Validate a given temperature hold time.

        Args:
            hold_time: The requested hold time in seconds.

        Raises:
            InvalidHoldTimeError: The given time is invalid

        Returns:
            The validated time in seconds
        """
        if hold_time < 0:
            raise InvalidHoldTimeError(
                "Thermocycler target temperature hold time must be a positive number,"
                f" but received {hold_time}."
            )
        return hold_time

    @staticmethod
    def validate_target_lid_temperature(celsius: float) -> float:
        """Validate a given target lid temperature.

        Args:
            celsius: The requested lid temperature.

        Raises:
            InvalidTargetTemperatureError: The given temperature
                is outside the thermocycler's operating range.

        Returns:
            The validated temperature in degrees Celsius.
        """
        if LID_TARGET_MIN <= celsius <= LID_TARGET_MAX:
            return celsius

        raise InvalidTargetTemperatureError(
            "Thermocycler lid temperature must be between"
            f" {LID_TARGET_MIN} and {LID_TARGET_MAX}, but got {celsius}."
        )

    def get_target_block_temperature(self) -> float:
        """Get the thermocycler's target block temperature."""
        target = self.target_block_temperature

        if target is None:
            raise NoTargetTemperatureSetError(
                f"Module {self.module_id} does not have a target block temperature set."
            )
        return target

    def get_target_lid_temperature(self) -> float:
        """Get the thermocycler's target lid temperature."""
        target = self.target_lid_temperature

        if target is None:
            raise NoTargetTemperatureSetError(
                f"Module {self.module_id} does not have a target block temperature set."
            )
        return target

    @classmethod
    def from_live_data(
        cls, module_id: ThermocyclerModuleId, data: ModuleData | None
    ) -> "ThermocyclerModuleSubState":
        """Create a ThermocyclerModuleSubState from live data."""
        if ModuleDataValidator.is_thermocycler_data(data):
            return cls(
                module_id=module_id,
                is_lid_open=data["lid"] == "open",
                target_block_temperature=data["targetTemp"],
                target_lid_temperature=data["lidTarget"],
            )
        return cls(
            module_id=module_id,
            is_lid_open=False,
            target_block_temperature=None,
            target_lid_temperature=None,
        )



import json
import logging
from pydantic import ValidationError
from typing import Optional, Union
from dataclasses import asdict

from opentrons import config
from opentrons.util.helpers import utc_now


from .. import file_operators as io, types as local_types

from .models import v1

log = logging.getLogger(__name__)

# Delete Deck Calibration


def delete_robot_deck_attitude() -> None:
    """
    Delete the robot deck attitude calibration.
    """
    legacy_deck_calibration_file = config.get_opentrons_path("deck_calibration_file")

    robot_dir = config.get_opentrons_path("robot_calibration_dir")
    gantry_path = robot_dir / "deck_calibration.json"

    # TODO(mc, 2022-06-08): this leaves legacy deck calibration backup files in place
    # we should eventually clean them up, too, because they can really crowd /data/
    io.delete_file(legacy_deck_calibration_file)
    io.delete_file(gantry_path)


# Save Deck Calibration


def save_robot_deck_attitude(
    transform: local_types.AttitudeMatrix,
    pip_id: Optional[str],
    lw_hash: Optional[str],
    source: Optional[local_types.SourceType] = None,
    cal_status: Optional[
        Union[local_types.CalibrationStatus, v1.CalibrationStatus]
    ] = None,
) -> None:
    robot_dir = config.get_opentrons_path("robot_calibration_dir")

    if isinstance(cal_status, local_types.CalibrationStatus):
        cal_status_model = v1.CalibrationStatus(**asdict(cal_status))
    elif isinstance(cal_status, v1.CalibrationStatus):
        cal_status_model = cal_status
    else:
        cal_status_model = v1.CalibrationStatus()

    gantry_calibration = v1.DeckCalibrationModel(
        attitude=transform,
        pipette_calibrated_with=pip_id,
        last_modified=utc_now(),
        tiprack=lw_hash,
        source=source or local_types.SourceType.user,
        status=cal_status_model,
    )
    # convert to schema + validate json conversion
    io.save_to_file(robot_dir, "deck_calibration", gantry_calibration)


# Get Deck Calibration


def get_robot_deck_attitude() -> Optional[v1.DeckCalibrationModel]:
    deck_calibration_path = (
        config.get_opentrons_path("robot_calibration_dir") / "deck_calibration.json"
    )
    try:
        return v1.DeckCalibrationModel(**io.read_cal_file(deck_calibration_path))
    except FileNotFoundError:
        log.warning("Deck calibration not found.")
        pass
    except (json.JSONDecodeError, ValidationError):
        log.warning(
            "Deck calibration is malformed. Please factory reset your calibrations.",
            exc_info=True,
        )
    return None



"""Move relative (jog) command payload, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal


from ..types import MovementAxis
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from .movement_common import (
    DestinationPositionResult,
    move_relative,
    StallOrCollisionError,
)

if TYPE_CHECKING:
    from ..execution import MovementHandler
    from ..resources.model_utils import ModelUtils


MoveRelativeCommandType = Literal["moveRelative"]


class MoveRelativeParams(BaseModel):
    """Payload required for a MoveRelative command."""

    pipetteId: str = Field(..., description="Pipette to move.")
    axis: MovementAxis = Field(..., description="Axis along which to move.")
    distance: float = Field(
        ...,
        description=(
            "Distance to move in millimeters. A positive number will move"
            " towards the right (x), back (y), top (z) of the deck."
        ),
    )


class MoveRelativeResult(DestinationPositionResult):
    """Result data from the execution of a MoveRelative command."""

    pass


class MoveRelativeImplementation(
    AbstractCommandImpl[
        MoveRelativeParams,
        SuccessData[MoveRelativeResult] | DefinedErrorData[StallOrCollisionError],
    ]
):
    """Move relative command implementation."""

    def __init__(
        self, movement: MovementHandler, model_utils: ModelUtils, **kwargs: object
    ) -> None:
        self._movement = movement
        self._model_utils = model_utils

    async def execute(
        self, params: MoveRelativeParams
    ) -> SuccessData[MoveRelativeResult] | DefinedErrorData[StallOrCollisionError]:
        """Move (jog) a given pipette a relative distance."""
        result = await move_relative(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=params.pipetteId,
            axis=params.axis,
            distance=params.distance,
        )
        if isinstance(result, DefinedErrorData):
            return result
        else:
            return SuccessData(
                public=MoveRelativeResult(position=result.public.position),
                state_update=result.state_update,
            )


class MoveRelative(
    BaseCommand[MoveRelativeParams, MoveRelativeResult, StallOrCollisionError]
):
    """Command to move (jog) a given pipette a relative distance."""

    commandType: MoveRelativeCommandType = "moveRelative"
    params: MoveRelativeParams
    result: Optional[MoveRelativeResult] = None

    _ImplementationCls: Type[MoveRelativeImplementation] = MoveRelativeImplementation


class MoveRelativeCreate(BaseCommandCreate[MoveRelativeParams]):
    """Data to create a MoveRelative command."""

    commandType: MoveRelativeCommandType = "moveRelative"
    params: MoveRelativeParams

    _CommandCls: Type[MoveRelative] = MoveRelative



"""Control a `ProtocolEngine` without async/await."""

from typing import cast, Any, Optional, overload

from opentrons_shared_data.labware.types import LabwareUri
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from .. import commands
from ..commands.command_unions import CREATE_TYPES_BY_PARAMS_TYPE
from ..state.state import StateView
from ..types import (
    Liquid,
    LabwareOffsetCreate,
)
from .transports import ChildThreadTransport


class SyncClient:
    """Control a `ProtocolEngine` without async/await.

    Normally, `ProtocolEngine` provides an async/await interface, like this:

    ```
    aspirate_result = await protocol_engine.add_and_execute_command(aspirate_command)
    dispense_result = await protocol_engine.add_and_execute_command(dispense_command)
    ```

    But we sometimes want to control it with plain old non-async blocking method calls.
    To accomplish that, this class adapts `ProtocolEngine`'s interface into this:

    ```
    aspirate_result = sync_client.aspirate(...)
    dispense_result = sync_client.dispense(...)
    ```

    This is intended to help implement the Python Protocol API, which is all non-async.
    """

    # todo(mm, 2024-06-13): The separation of responsibilities between this class
    # and ChildThreadTransport has grown to be pretty fuzzy. Clarify or merge the two
    # classes.

    def __init__(self, transport: ChildThreadTransport) -> None:
        """Initialize the `SyncClient`.

        Params:
            transport: The interface for the new `SyncClient` to use to
                communicate with the `ProtocolEngine`.
        """
        self._transport = transport

    def execute_command(self, params: commands.CommandParams) -> None:
        """Execute a ProtocolEngine command, including error recovery.

        See `ChildThreadTransport.execute_command_wait_for_recovery()` for exact
        behavior.
        """
        CreateType = CREATE_TYPES_BY_PARAMS_TYPE[type(params)]
        create_request = CreateType(params=cast(Any, params))
        self._transport.execute_command_wait_for_recovery(create_request)

    @overload
    def execute_command_without_recovery(
        self, params: commands.LoadLabwareParams
    ) -> commands.LoadLabwareResult:
        pass

    @overload
    def execute_command_without_recovery(
        self, params: commands.LoadModuleParams
    ) -> commands.LoadModuleResult:
        pass

    @overload
    def execute_command_without_recovery(
        self, params: commands.LoadPipetteParams
    ) -> commands.LoadPipetteResult:
        pass

    @overload
    def execute_command_without_recovery(
        self, params: commands.LoadLidStackParams
    ) -> commands.LoadLidStackResult:
        pass

    @overload
    def execute_command_without_recovery(
        self, params: commands.LoadLidParams
    ) -> commands.LoadLidResult:
        pass

    @overload
    def execute_command_without_recovery(
        self, params: commands.LiquidProbeParams
    ) -> commands.LiquidProbeResult:
        pass

    @overload
    def execute_command_without_recovery(
        self, params: commands.TryLiquidProbeParams
    ) -> commands.TryLiquidProbeResult:
        pass

    @overload
    def execute_command_without_recovery(
        self, params: commands.LoadLiquidClassParams
    ) -> commands.LoadLiquidClassResult:
        pass

    @overload
    def execute_command_without_recovery(
        self, params: commands.GetNextTipParams
    ) -> commands.GetNextTipResult:
        pass

    def execute_command_without_recovery(
        self, params: commands.CommandParams
    ) -> commands.CommandResult:
        """Execute a ProtocolEngine command.

        See `ChildThreadTransport.execute_command()` for exact
        behavior.
        """
        CreateType = CREATE_TYPES_BY_PARAMS_TYPE[type(params)]
        create_request = CreateType(params=cast(Any, params))
        return self._transport.execute_command(create_request)

    @property
    def state(self) -> StateView:
        """Get a view of the engine's state."""
        return self._transport.state

    def add_labware_definition(self, definition: LabwareDefinition) -> LabwareUri:
        """Add a labware definition to the engine."""
        return self._transport.call_method(
            "add_labware_definition",
            definition=definition,
        )

    def add_addressable_area(self, addressable_area_name: str) -> None:
        """Add an addressable area to the engine's state."""
        self._transport.call_method(
            "add_addressable_area", addressable_area_name=addressable_area_name
        )

    def add_liquid(
        self, name: str, color: Optional[str], description: Optional[str]
    ) -> Liquid:
        """Add a liquid to the engine."""
        return self._transport.call_method("add_liquid", name=name, color=color, description=description)  # type: ignore[no-any-return]

    def reset_tips(self, labware_id: str) -> None:
        """Reset a labware's tip tracking state.."""
        self._transport.call_method(
            "reset_tips",
            labware_id=labware_id,
        )

    def add_labware_offset(self, request: LabwareOffsetCreate) -> None:
        """Add a labware offset."""
        self._transport.call_method("add_labware_offset", request=request)

    def set_pipette_movement_speed(
        self, pipette_id: str, speed: Optional[float]
    ) -> None:
        """Set the speed of a pipette's X/Y/Z movements. Does not affect plunger speed.

        None will use the hardware API's default.
        """
        self._transport.call_method(
            "set_pipette_movement_speed",
            pipette_id=pipette_id,
            speed=speed,
        )



"""Command models for Absorbance Reader commands."""
from .close_lid import (
    CloseLidCommandType,
    CloseLidParams,
    CloseLidResult,
    CloseLid,
    CloseLidCreate,
)

from .open_lid import (
    OpenLidCommandType,
    OpenLidParams,
    OpenLidResult,
    OpenLid,
    OpenLidCreate,
)

from .initialize import (
    InitializeCommandType,
    InitializeParams,
    InitializeResult,
    Initialize,
    InitializeCreate,
)

from .read import (
    ReadAbsorbanceCommandType,
    ReadAbsorbanceParams,
    ReadAbsorbanceResult,
    ReadAbsorbance,
    ReadAbsorbanceCreate,
)


__all__ = [
    # absorbanace_reader/closeLid
    "CloseLidCommandType",
    "CloseLidParams",
    "CloseLidResult",
    "CloseLid",
    "CloseLidCreate",
    # absorbanace_reader/openLid
    "OpenLidCommandType",
    "OpenLidParams",
    "OpenLidResult",
    "OpenLid",
    "OpenLidCreate",
    # absorbanace_reader/initialize
    "InitializeCommandType",
    "InitializeParams",
    "InitializeResult",
    "Initialize",
    "InitializeCreate",
    # absorbanace_reader/measure
    "ReadAbsorbanceCommandType",
    "ReadAbsorbanceParams",
    "ReadAbsorbanceResult",
    "ReadAbsorbance",
    "ReadAbsorbanceCreate",
    # union type
]



from __future__ import annotations
from enum import Enum
from dataclasses import dataclass
from typing import (
    Dict,
    List,
    NamedTuple,
    Callable,
    Any,
    Tuple,
    Awaitable,
    Union,
    Optional,
    cast,
    TYPE_CHECKING,
    TypeGuard,
)
from typing_extensions import TypedDict
from pathlib import Path

from opentrons.drivers.flex_stacker.types import (
    LimitSwitchStatus,
    PlatformStatus,
    StackerAxis,
)
from opentrons.drivers.rpi_drivers.types import USBPort

if TYPE_CHECKING:
    from opentrons_shared_data.module.types import (
        ThermocyclerModuleType,
        MagneticModuleType,
        TemperatureModuleType,
        HeaterShakerModuleType,
        MagneticBlockType,
        AbsorbanceReaderType,
        FlexStackerModuleType,
    )


class ThermocyclerStepBase(TypedDict):
    temperature: float


class ThermocyclerStep(ThermocyclerStepBase, total=False):
    hold_time_seconds: float
    hold_time_minutes: float


class ThermocyclerCycle(TypedDict):
    steps: List[ThermocyclerStep]
    repetitions: int


UploadFunction = Callable[[str, str, Dict[str, Any]], Awaitable[Tuple[bool, str]]]


ModuleDisconnectedCallback = Optional[Callable[[str, str | None], None]]


class MagneticModuleData(TypedDict):
    engaged: bool
    height: float


class TemperatureModuleData(TypedDict):
    currentTemp: float
    targetTemp: float | None


class HeaterShakerData(TypedDict):
    temperatureStatus: str
    speedStatus: str
    labwareLatchStatus: str
    currentTemp: float
    targetTemp: float | None
    currentSpeed: int
    targetSpeed: int | None
    errorDetails: str | None


class ThermocyclerData(TypedDict):
    lid: str
    lidTarget: float | None
    lidTemp: float
    lidTempStatus: str
    currentTemp: float | None
    targetTemp: float | None
    holdTime: float | None
    rampRate: float | None
    currentCycleIndex: int | None
    totalCycleCount: int | None
    currentStepIndex: int | None
    totalStepCount: int | None


class AbsorbanceReaderData(TypedDict):
    uptime: int
    deviceStatus: str
    lidStatus: str
    platePresence: str
    measureMode: str
    sampleWavelengths: List[int]
    referenceWavelength: int


class FlexStackerData(TypedDict):
    latchState: str
    platformState: str
    hopperDoorState: str
    axisStateX: str
    axisStateZ: str
    errorDetails: str | None


ModuleData = Union[
    Dict[Any, Any],  # This allows an empty dict as module data
    MagneticModuleData,
    TemperatureModuleData,
    HeaterShakerData,
    ThermocyclerData,
    AbsorbanceReaderData,
    FlexStackerData,
]


class ModuleDataValidator:
    @classmethod
    def is_magnetic_module_data(
        cls, data: ModuleData | None
    ) -> TypeGuard[MagneticModuleData]:
        return data is not None and "engaged" in data.keys()

    @classmethod
    def is_temperature_module_data(
        cls, data: ModuleData | None
    ) -> TypeGuard[TemperatureModuleData]:
        return data is not None and "targetTemp" in data.keys()

    @classmethod
    def is_heater_shaker_data(
        cls, data: ModuleData | None
    ) -> TypeGuard[HeaterShakerData]:
        return data is not None and "labwareLatchStatus" in data.keys()

    @classmethod
    def is_thermocycler_data(
        cls, data: ModuleData | None
    ) -> TypeGuard[ThermocyclerData]:
        return data is not None and "lid" in data.keys()

    @classmethod
    def is_absorbance_reader_data(
        cls, data: ModuleData | None
    ) -> TypeGuard[AbsorbanceReaderData]:
        return data is not None and "uptime" in data.keys()

    @classmethod
    def is_flex_stacker_data(
        cls, data: ModuleData | None
    ) -> TypeGuard[FlexStackerData]:
        return data is not None and "platformState" in data.keys()


class LiveData(TypedDict):
    status: str
    data: ModuleData | None


class ModuleType(str, Enum):
    THERMOCYCLER: ThermocyclerModuleType = "thermocyclerModuleType"
    TEMPERATURE: TemperatureModuleType = "temperatureModuleType"
    MAGNETIC: MagneticModuleType = "magneticModuleType"
    HEATER_SHAKER: HeaterShakerModuleType = "heaterShakerModuleType"
    MAGNETIC_BLOCK: MagneticBlockType = "magneticBlockType"
    ABSORBANCE_READER: AbsorbanceReaderType = "absorbanceReaderType"
    FLEX_STACKER: FlexStackerModuleType = "flexStackerModuleType"

    @classmethod
    def from_model(cls, model: ModuleModel) -> ModuleType:
        if isinstance(model, MagneticModuleModel):
            return cls.MAGNETIC
        if isinstance(model, TemperatureModuleModel):
            return cls.TEMPERATURE
        if isinstance(model, ThermocyclerModuleModel):
            return cls.THERMOCYCLER
        if isinstance(model, HeaterShakerModuleModel):
            return cls.HEATER_SHAKER
        if isinstance(model, MagneticBlockModel):
            return cls.MAGNETIC_BLOCK
        if isinstance(model, AbsorbanceReaderModel):
            return cls.ABSORBANCE_READER
        if isinstance(model, FlexStackerModuleModel):
            return cls.FLEX_STACKER

    @classmethod
    def to_module_fixture_id(cls, module_type: ModuleType) -> str:
        if module_type == ModuleType.THERMOCYCLER:
            # Thermocyclers are "loaded" in B1 only
            return "thermocyclerModuleV2Front"
        if module_type == ModuleType.TEMPERATURE:
            return "temperatureModuleV2"
        if module_type == ModuleType.HEATER_SHAKER:
            return "heaterShakerModuleV1"
        if module_type == ModuleType.MAGNETIC_BLOCK:
            return "magneticBlockV1"
        if module_type == ModuleType.ABSORBANCE_READER:
            return "absorbanceReaderV1"
        if module_type == ModuleType.FLEX_STACKER:
            return "flexStackerModuleV1"
        else:
            raise ValueError(
                f"Module Type {module_type} does not have a related fixture ID."
            )


class MagneticModuleModel(str, Enum):
    MAGNETIC_V1: str = "magneticModuleV1"
    MAGNETIC_V2: str = "magneticModuleV2"


class TemperatureModuleModel(str, Enum):
    TEMPERATURE_V1: str = "temperatureModuleV1"
    TEMPERATURE_V2: str = "temperatureModuleV2"


class ThermocyclerModuleModel(str, Enum):
    THERMOCYCLER_V1: str = "thermocyclerModuleV1"
    THERMOCYCLER_V2: str = "thermocyclerModuleV2"


class HeaterShakerModuleModel(str, Enum):
    HEATER_SHAKER_V1: str = "heaterShakerModuleV1"


class MagneticBlockModel(str, Enum):
    MAGNETIC_BLOCK_V1: str = "magneticBlockV1"


class AbsorbanceReaderModel(str, Enum):
    ABSORBANCE_READER_V1: str = "absorbanceReaderV1"


class FlexStackerModuleModel(str, Enum):
    FLEX_STACKER_V1: str = "flexStackerModuleV1"


def module_model_from_string(model_string: str) -> ModuleModel:
    for model_enum in {
        MagneticModuleModel,
        TemperatureModuleModel,
        ThermocyclerModuleModel,
        HeaterShakerModuleModel,
        MagneticBlockModel,
        AbsorbanceReaderModel,
        FlexStackerModuleModel,
    }:
        try:
            return cast(ModuleModel, model_enum(model_string))
        except ValueError:
            pass
    raise ValueError(f"No such module model {model_string}")


@dataclass(kw_only=True)
class ModuleAtPort:
    port: str
    name: str
    serial: Optional[str] = None
    usb_port: USBPort = USBPort(name="", port_number=0)


@dataclass(kw_only=True)
class SimulatingModule:
    serial_number: str
    model: Optional[str]


@dataclass(kw_only=True)
class SimulatingModuleAtPort(ModuleAtPort, SimulatingModule):
    pass


class BundledFirmware(NamedTuple):
    """Represents a versioned firmware file, generally bundled into the fs"""

    version: str
    path: Path

    def __repr__(self) -> str:
        return f"<BundledFirmware {self.version}, path={self.path}>"


class ModuleInfo(NamedTuple):
    model: str  # A module model such as "magneticModuleV2"
    fw_version: str  # The version of the firmware
    hw_revision: str  # the revision of the hardware
    serial: str  # the serial number


# TODO(mc, 2022-01-18): replace with enum
ModuleModel = Union[
    MagneticModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
    HeaterShakerModuleModel,
    MagneticBlockModel,
    AbsorbanceReaderModel,
    FlexStackerModuleModel,
]


class MagneticStatus(str, Enum):
    ENGAGED = "engaged"
    DISENGAGED = "disengaged"


class TemperatureStatus(str, Enum):
    HOLDING = "holding at target"
    COOLING = "cooling"
    HEATING = "heating"
    IDLE = "idle"
    ERROR = "error"


class SpeedStatus(str, Enum):
    HOLDING = "holding at target"
    ACCELERATING = "speeding up"
    DECELERATING = "slowing down"
    IDLE = "idle"
    ERROR = "error"


class HeaterShakerStatus(str, Enum):
    IDLE = "idle"
    RUNNING = "running"
    ERROR = "error"


class AbsorbanceReaderStatus(str, Enum):
    IDLE = "idle"
    MEASURING = "measuring"
    ERROR = "error"


class LidStatus(str, Enum):
    ON = "on"
    OFF = "off"
    UNKNOWN = "unknown"
    ERROR = "error"


class FlexStackerStatus(str, Enum):
    IDLE = "idle"
    DISPENSING = "dispensing"
    STORING = "storing"
    ERROR = "error"


class PlatformState(str, Enum):
    UNKNOWN = "unknown"
    EXTENDED = "extended"
    RETRACTED = "retracted"

    @classmethod
    def from_status(cls, status: PlatformStatus) -> "PlatformState":
        """Get the state from the platform status."""
        if status.E and not status.R:
            return cls.EXTENDED
        if status.R and not status.E:
            return cls.RETRACTED
        return cls.UNKNOWN


class StackerAxisState(str, Enum):
    UNKNOWN = "unknown"
    EXTENDED = "extended"
    RETRACTED = "retracted"

    @classmethod
    def from_status(
        cls, status: LimitSwitchStatus, axis: StackerAxis
    ) -> "StackerAxisState":
        """Get the axis state from the limit switch status."""
        match axis:
            case StackerAxis.X:
                if status.XE and not status.XR:
                    return cls.EXTENDED
                if status.XR and not status.XE:
                    return cls.RETRACTED
            case StackerAxis.Z:
                if status.ZE and not status.ZR:
                    return cls.EXTENDED
                if status.ZR and not status.ZE:
                    return cls.RETRACTED
            case StackerAxis.L:
                return cls.EXTENDED if status.LR else cls.RETRACTED
        return cls.UNKNOWN


class LatchState(str, Enum):
    CLOSED = "closed"
    OPENED = "opened"

    @classmethod
    def from_state(cls, state: StackerAxisState) -> "LatchState":
        """Get the latch state from the axis state."""
        return cls.CLOSED if state == StackerAxisState.EXTENDED else cls.OPENED


class HopperDoorState(str, Enum):
    CLOSED = "closed"
    OPENED = "opened"

    @classmethod
    def from_state(cls, state: bool) -> "HopperDoorState":
        """Get the hopper door state from the door state boolean."""
        return cls.CLOSED if state else cls.OPENED



"""The interface that implements InstrumentContext."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, List, NamedTuple, Optional, TypeVar, Dict

from opentrons_shared_data.labware.types import (
    LabwareUri,
    LabwareParameters2,
    LabwareParameters3,
    LabwareDefinition as LabwareDefinitionDict,
)

from opentrons.types import DeckSlotName, Point, NozzleMapInterface
from .._liquid import Liquid

from .well import WellCoreType


_LabwareParametersDict = LabwareParameters2 | LabwareParameters3


class LabwareLoadParams(NamedTuple):
    """Unique load parameters of a labware."""

    namespace: str
    load_name: str
    version: int

    def as_uri(self) -> LabwareUri:
        """Get the labware's definition URI from the load parameters."""
        return LabwareUri(f"{self.namespace}/{self.load_name}/{self.version}")

    @classmethod
    def from_uri(cls, uri: LabwareUri) -> LabwareLoadParams:
        namespace, load_name, version_str = uri.split("/")
        return cls(namespace, load_name, int(version_str))


class AbstractLabware(ABC, Generic[WellCoreType]):
    """Labware implementation core interface."""

    @property
    @abstractmethod
    def load_name(self) -> str:
        """Get the labware's load name."""

    @property
    @abstractmethod
    def highest_z(self) -> float:
        """Get the Z coordinate of the labware's highest point"""

    @abstractmethod
    def get_uri(self) -> str:
        """Get the URI string string of the labware's definition.

        The URI is unique for a given namespace, load name, and definition version.
        """

    @abstractmethod
    def get_load_params(self) -> LabwareLoadParams:
        ...

    @abstractmethod
    def get_display_name(self) -> str:
        """Get a display name for the labware, falling back to the definition."""

    @abstractmethod
    def get_user_display_name(self) -> Optional[str]:
        """Get the user-specified display name of the labware, if set."""

    @abstractmethod
    def get_name(self) -> str:
        ...

    @abstractmethod
    def get_definition(self) -> LabwareDefinitionDict:
        """Get the labware's definition as a plain dictionary."""

    @abstractmethod
    def get_parameters(self) -> _LabwareParametersDict:
        """Get the labware's definition's `parameters` field as a plain dictionary."""

    @abstractmethod
    def get_quirks(self) -> List[str]:
        ...

    @abstractmethod
    def set_calibration(self, delta: Point) -> None:
        ...

    @abstractmethod
    def get_calibrated_offset(self) -> Point:
        ...

    @abstractmethod
    def is_tip_rack(self) -> bool:
        """Whether the labware is a tip rack."""

    @abstractmethod
    def is_adapter(self) -> bool:
        """Whether the labware is an adapter."""

    @abstractmethod
    def is_lid(self) -> bool:
        """Whether the labware is a lid."""

    @abstractmethod
    def is_fixed_trash(self) -> bool:
        """Whether the labware is a fixed trash."""

    @abstractmethod
    def get_tip_length(self) -> float:
        ...

    @abstractmethod
    def reset_tips(self) -> None:
        ...

    @abstractmethod
    def get_next_tip(
        self,
        num_tips: int,
        starting_tip: Optional[WellCoreType],
        nozzle_map: Optional[NozzleMapInterface],
    ) -> Optional[str]:
        """Get the name of the next available tip(s) in the rack, if available."""

    @abstractmethod
    def get_well_columns(self) -> List[List[str]]:
        """Get the all well names, organized by column, from the labware's definition."""

    @abstractmethod
    def get_well_core(self, well_name: str) -> WellCoreType:
        """Get a well core interface to a given well in this labware."""

    @abstractmethod
    def get_deck_slot(self) -> Optional[DeckSlotName]:
        """Get the deck slot the labware or its parent is in, if any."""

    @abstractmethod
    def load_liquid(self, volumes: Dict[str, float], liquid: Liquid) -> None:
        """Load liquid into wells of the labware."""

    @abstractmethod
    def load_empty(self, wells: List[str]) -> None:
        """Mark wells of the labware as empty."""


LabwareCoreType = TypeVar("LabwareCoreType", bound=AbstractLabware[Any])



"""Command models to open the Heater-Shaker Module's labware latch."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type
from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...state import update_types

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler, MovementHandler

OpenLabwareLatchCommandType = Literal["heaterShaker/openLabwareLatch"]


class OpenLabwareLatchParams(BaseModel):
    """Input parameters to open a Heater-Shaker Module's labware latch."""

    moduleId: str = Field(..., description="Unique ID of the Heater-Shaker Module.")


class OpenLabwareLatchResult(BaseModel):
    """Result data from opening a Heater-Shaker's labware latch."""

    pipetteRetracted: bool = Field(
        ...,
        description=(
            "Whether this command automatically retracted the pipettes"
            " before opening the latch, to avoid a potential collision."
        ),
    )


class OpenLabwareLatchImpl(
    AbstractCommandImpl[OpenLabwareLatchParams, SuccessData[OpenLabwareLatchResult]]
):
    """Execution implementation of a Heater-Shaker's open latch labware command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        movement: MovementHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._movement = movement

    async def execute(
        self, params: OpenLabwareLatchParams
    ) -> SuccessData[OpenLabwareLatchResult]:
        """Open a Heater-Shaker's labware latch."""
        state_update = update_types.StateUpdate()

        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        hs_module_substate = self._state_view.modules.get_heater_shaker_module_substate(
            module_id=params.moduleId
        )

        hs_module_substate.raise_if_shaking()

        pipette_should_retract = (
            self._state_view.motion.check_pipette_blocking_hs_latch(
                hs_module_substate.module_id
            )
        )
        if pipette_should_retract:
            # Move pipette away if it is close to the heater-shaker
            # TODO(jbl 2022-07-28) replace home movement with a retract movement
            await self._movement.home(
                axes=self._state_view.motion.get_robot_mount_axes()
            )
            state_update.clear_all_pipette_locations()

        # Allow propagation of ModuleNotAttachedError.
        hs_hardware_module = self._equipment.get_module_hardware_api(
            hs_module_substate.module_id
        )

        if hs_hardware_module is not None:
            await hs_hardware_module.open_labware_latch()

        return SuccessData(
            public=OpenLabwareLatchResult(pipetteRetracted=pipette_should_retract),
            state_update=state_update,
        )


class OpenLabwareLatch(
    BaseCommand[OpenLabwareLatchParams, OpenLabwareLatchResult, ErrorOccurrence]
):
    """A command to open a Heater-Shaker's labware latch."""

    commandType: OpenLabwareLatchCommandType = "heaterShaker/openLabwareLatch"
    params: OpenLabwareLatchParams
    result: Optional[OpenLabwareLatchResult] = None

    _ImplementationCls: Type[OpenLabwareLatchImpl] = OpenLabwareLatchImpl


class OpenLabwareLatchCreate(BaseCommandCreate[OpenLabwareLatchParams]):
    """A request to create a Heater-Shaker's open labware latch command."""

    commandType: OpenLabwareLatchCommandType = "heaterShaker/openLabwareLatch"
    params: OpenLabwareLatchParams

    _CommandCls: Type[OpenLabwareLatch] = OpenLabwareLatch



"""Aspirate command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal

from .pipetting_common import (
    OverpressureError,
    PipetteIdMixin,
    AspirateVolumeMixin,
    FlowRateMixin,
    BaseLiquidHandlingResult,
    aspirate_while_tracking,
)
from .movement_common import (
    LiquidHandlingWellLocationMixin,
    DestinationPositionResult,
    StallOrCollisionError,
    move_to_well,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..state.update_types import StateUpdate
from ..errors.exceptions import PipetteNotReadyToAspirateError
from opentrons.hardware_control import HardwareControlAPI
from ..state.update_types import CLEAR
from ..types import DeckPoint

if TYPE_CHECKING:
    from ..execution import PipettingHandler, GantryMover, MovementHandler
    from ..resources import ModelUtils
    from ..state.state import StateView
    from ..notes import CommandNoteAdder


AspirateWhileTrackingCommandType = Literal["aspirateWhileTracking"]


class AspirateWhileTrackingParams(
    PipetteIdMixin,
    AspirateVolumeMixin,
    FlowRateMixin,
    LiquidHandlingWellLocationMixin,
):
    """Parameters required to aspirate from a specific well."""

    pass


class AspirateWhileTrackingResult(BaseLiquidHandlingResult, DestinationPositionResult):
    """Result data from execution of an Aspirate command."""

    pass


_ExecuteReturn = Union[
    SuccessData[AspirateWhileTrackingResult],
    DefinedErrorData[OverpressureError] | DefinedErrorData[StallOrCollisionError],
]


class AspirateWhileTrackingImplementation(
    AbstractCommandImpl[AspirateWhileTrackingParams, _ExecuteReturn]
):
    """AspirateWhileTracking command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        command_note_adder: CommandNoteAdder,
        model_utils: ModelUtils,
        gantry_mover: GantryMover,
        movement: MovementHandler,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api
        self._command_note_adder = command_note_adder
        self._model_utils = model_utils
        self._gantry_mover = gantry_mover
        self._movement = movement

    async def execute(self, params: AspirateWhileTrackingParams) -> _ExecuteReturn:
        """Move to and aspirate from the requested well.

        Raises:
            TipNotAttachedError: if no tip is attached to the pipette.
            PipetteNotReadyToAspirateError: pipette plunger is not ready.
        """
        ready_to_aspirate = self._state_view.pipettes.get_ready_to_aspirate(
            pipette_id=params.pipetteId
        )
        if not ready_to_aspirate:
            raise PipetteNotReadyToAspirateError(
                "Pipette cannot aspirate while tracking because of a previous blow out."
                " The first aspirate following a blow-out must be from a specific well"
                " so the plunger can be reset in a known safe position."
            )
        state_update = StateUpdate()

        move_result = await move_to_well(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=params.pipetteId,
            labware_id=params.labwareId,
            well_name=params.wellName,
            well_location=params.wellLocation,
            operation_volume=-params.volume,
        )
        state_update.append(move_result.state_update)
        if isinstance(move_result, DefinedErrorData):
            return DefinedErrorData(
                public=move_result.public, state_update=state_update
            )

        aspirate_result = await aspirate_while_tracking(
            pipette_id=params.pipetteId,
            labware_id=params.labwareId,
            well_name=params.wellName,
            volume=params.volume,
            flow_rate=params.flowRate,
            location_if_error={
                "retryLocation": (
                    move_result.public.position.x,
                    move_result.public.position.y,
                    move_result.public.position.z,
                )
            },
            command_note_adder=self._command_note_adder,
            pipetting=self._pipetting,
            model_utils=self._model_utils,
        )
        position_after_aspirate = await self._gantry_mover.get_position(
            params.pipetteId
        )
        result_deck_point = DeckPoint.model_construct(
            x=position_after_aspirate.x,
            y=position_after_aspirate.y,
            z=position_after_aspirate.z,
        )
        if isinstance(aspirate_result, DefinedErrorData):
            return DefinedErrorData(
                public=aspirate_result.public,
                state_update=aspirate_result.state_update.set_liquid_operated(
                    labware_id=params.labwareId,
                    well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                        params.labwareId,
                        params.wellName,
                        params.pipetteId,
                    ),
                    volume_added=CLEAR,
                ),
                state_update_if_false_positive=aspirate_result.state_update_if_false_positive,
            )

        return SuccessData(
            public=AspirateWhileTrackingResult(
                volume=aspirate_result.public.volume,
                position=result_deck_point,
            ),
            state_update=aspirate_result.state_update.set_liquid_operated(
                labware_id=params.labwareId,
                well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                    params.labwareId,
                    params.wellName,
                    params.pipetteId,
                ),
                volume_added=-aspirate_result.public.volume
                * self._state_view.geometry.get_nozzles_per_well(
                    params.labwareId,
                    params.wellName,
                    params.pipetteId,
                ),
            ),
        )


class AspirateWhileTracking(
    BaseCommand[
        AspirateWhileTrackingParams,
        AspirateWhileTrackingResult,
        OverpressureError | StallOrCollisionError,
    ]
):
    """AspirateWhileTracking command model."""

    commandType: AspirateWhileTrackingCommandType = "aspirateWhileTracking"
    params: AspirateWhileTrackingParams
    result: Optional[AspirateWhileTrackingResult] = None

    _ImplementationCls: Type[
        AspirateWhileTrackingImplementation
    ] = AspirateWhileTrackingImplementation


class AspirateWhileTrackingCreate(BaseCommandCreate[AspirateWhileTrackingParams]):
    """Create aspirateWhileTracking command request model."""

    commandType: AspirateWhileTrackingCommandType = "aspirateWhileTracking"
    params: AspirateWhileTrackingParams

    _CommandCls: Type[AspirateWhileTracking] = AspirateWhileTracking



from enum import Enum
from dataclasses import dataclass, asdict, fields
from typing import Dict, Tuple, TypeVar, Generic, List, cast
from typing_extensions import TypedDict, Literal
from opentrons.hardware_control.types import OT3AxisKind


class AxisDict(TypedDict):
    X: float
    Y: float
    Z: float
    A: float
    B: float
    C: float


Vt = TypeVar("Vt")


class GantryLoad(Enum):
    HIGH_THROUGHPUT = "high_throughput"
    LOW_THROUGHPUT = "low_throughput"


@dataclass
class ByGantryLoad(Generic[Vt]):
    high_throughput: Vt
    low_throughput: Vt

    def __getitem__(self, key: GantryLoad) -> Vt:
        return cast(Vt, asdict(self)[key.value])


PerPipetteAxisSettings = ByGantryLoad[Dict[OT3AxisKind, float]]


class CurrentDictDefault(TypedDict):
    default: AxisDict


CurrentDictModelEntries = TypedDict(
    "CurrentDictModelEntries",
    {"2.1": AxisDict, "A": AxisDict, "B": AxisDict, "C": AxisDict},
    total=False,
)


class CurrentDict(CurrentDictDefault, CurrentDictModelEntries):
    pass


Offset = Tuple[float, float, float]


@dataclass
class RobotConfig:
    model: Literal["OT-2 Standard"]
    name: str
    version: int
    gantry_steps_per_mm: Dict[str, float]
    acceleration: Dict[str, float]
    serial_speed: int
    default_pipette_configs: Dict[str, float]
    default_current: CurrentDict
    low_current: CurrentDict
    high_current: CurrentDict
    default_max_speed: AxisDict
    log_level: str
    z_retract_distance: float
    left_mount_offset: Offset


OT3Transform = List[List[float]]


@dataclass(frozen=True)
class OT3MotionSettings:
    default_max_speed: PerPipetteAxisSettings
    acceleration: PerPipetteAxisSettings
    max_speed_discontinuity: PerPipetteAxisSettings
    direction_change_speed_discontinuity: PerPipetteAxisSettings

    def by_gantry_load(
        self, gantry_load: GantryLoad
    ) -> Dict[str, Dict[OT3AxisKind, float]]:
        return dict(
            (field.name, getattr(self, field.name)[gantry_load])
            for field in fields(self)
        )


@dataclass(frozen=True)
class OT3CurrentSettings:
    hold_current: PerPipetteAxisSettings
    run_current: PerPipetteAxisSettings

    def by_gantry_load(
        self, gantry_load: GantryLoad
    ) -> Dict[str, Dict[OT3AxisKind, float]]:
        return dict(
            (field.name, getattr(self, field.name)[gantry_load])
            for field in fields(self)
        )


@dataclass(frozen=True)
class CapacitivePassSettings:
    prep_distance_mm: float
    max_overrun_distance_mm: float
    speed_mm_per_s: float
    sensor_threshold_pf: float


@dataclass(frozen=True)
class ZSenseSettings:
    pass_settings: CapacitivePassSettings


@dataclass
class LiquidProbeSettings:
    mount_speed: float
    plunger_speed: float
    plunger_impulse_time: float
    sensor_threshold_pascals: float
    aspirate_while_sensing: bool
    z_overlap_between_passes_mm: float
    plunger_reset_offset: float
    samples_for_baselining: int
    sample_time_sec: float


@dataclass(frozen=True)
class EdgeSenseSettings:
    overrun_tolerance_mm: float
    early_sense_tolerance_mm: float
    pass_settings: CapacitivePassSettings
    search_initial_tolerance_mm: float
    search_iteration_limit: int

    def __init__(
        self,
        overrun_tolerance_mm: float,
        early_sense_tolerance_mm: float,
        pass_settings: CapacitivePassSettings,
        search_initial_tolerance_mm: float,
        search_iteration_limit: int,
    ) -> None:
        if overrun_tolerance_mm > pass_settings.max_overrun_distance_mm:
            raise ValueError("Overrun tolerance and pass setting distance do not match")
        object.__setattr__(self, "overrun_tolerance_mm", overrun_tolerance_mm)
        object.__setattr__(self, "early_sense_tolerance_mm", early_sense_tolerance_mm)
        object.__setattr__(self, "pass_settings", pass_settings)
        object.__setattr__(
            self, "search_initial_tolerance_mm", search_initial_tolerance_mm
        )
        object.__setattr__(self, "search_iteration_limit", search_iteration_limit)


@dataclass(frozen=True)
class OT3CalibrationSettings:
    z_offset: ZSenseSettings
    edge_sense: EdgeSenseSettings
    probe_length: float


@dataclass
class OT3Config:
    model: Literal["OT-3 Standard"]
    name: str
    version: int
    log_level: str
    motion_settings: OT3MotionSettings
    current_settings: OT3CurrentSettings
    safe_home_distance: float
    deck_transform: OT3Transform
    carriage_offset: Offset
    left_mount_offset: Offset
    right_mount_offset: Offset
    gripper_mount_offset: Offset
    calibration: OT3CalibrationSettings
    liquid_sense: LiquidProbeSettings



"""Protocol Engine types dealing with command annotations."""

from typing import List, Literal, Dict, Any, Optional, Union
from pydantic import (
    ConfigDict,
    BaseModel,
    Field,
)


class BaseCommandAnnotation(BaseModel):
    """Optional annotations for protocol engine commands."""

    commandKeys: List[str] = Field(
        ..., description="Command keys to which this annotation applies"
    )
    annotationType: str = Field(
        ..., description="The type of annotation (for machine parsing)"
    )


class SecondOrderCommandAnnotation(BaseCommandAnnotation):
    """Annotates a group of atomic commands which were the direct result of a second order command.

    Examples of second order commands would be transfer, consolidate, mix, etc.
    """

    annotationType: Literal["secondOrderCommand"] = "secondOrderCommand"
    params: Dict[str, Any] = Field(
        ...,
        description="Key value pairs of the parameters passed to the second order command that this annotates.",
    )
    machineReadableName: str = Field(
        ...,
        description="The name of the second order command in the form that the generating software refers to it",
    )
    userSpecifiedName: Optional[str] = Field(
        None, description="The optional user-specified name of the second order command"
    )
    userSpecifiedDescription: Optional[str] = Field(
        None,
        description="The optional user-specified description of the second order command",
    )


class CustomCommandAnnotation(BaseCommandAnnotation):
    """Annotates a group of atomic commands in some manner that Opentrons software does not anticipate or originate."""

    annotationType: Literal["custom"] = "custom"
    model_config = ConfigDict(extra="allow")


CommandAnnotation = Union[SecondOrderCommandAnnotation, CustomCommandAnnotation]



"""Utility functions for transfer_liquid, consolidate_liquid and distribute_liquid"""
from __future__ import annotations

from typing import Literal, Sequence, List, Optional, TYPE_CHECKING
from dataclasses import dataclass

from opentrons.protocol_engine.errors import (
    LiquidHeightUnknownError,
    IncompleteLabwareDefinitionError,
)
from opentrons.protocol_engine.state._well_math import (
    wells_covered_by_pipette_configuration,
)
from opentrons.types import NozzleMapInterface, NozzleConfigurationType

if TYPE_CHECKING:
    from logging import Logger
    from opentrons.types import Location
    from opentrons.protocol_api.core.engine import WellCore
    from opentrons.protocol_api.labware import Well, Labware


@dataclass
class LocationCheckDescriptors:
    location_type: Literal["submerge start", "retract end"]
    pipetting_action: Literal["aspirate", "dispense"]


def raise_if_location_inside_liquid(
    location: Location,
    well_core: WellCore,
    location_check_descriptors: LocationCheckDescriptors,
    logger: Logger,
) -> None:
    """Raise an error if the location in question would be inside the liquid.

    This checker will raise an error if we can find the liquid height
    AND the location in question is below this height.

    If we can't find the liquid height, then we simply log the details and no error is raised.
    """
    try:
        liquid_height_from_bottom = well_core.current_liquid_height()
    except (IncompleteLabwareDefinitionError, LiquidHeightUnknownError):
        # IncompleteLabwareDefinitionError is raised when there's no inner geometry
        # defined for the well. So, we can't find the liquid height even if liquid volume is known.
        # LiquidHeightUnknownError is raised when we don't have liquid volume info
        # and no probing has been done either.
        liquid_height_from_bottom = None
    if isinstance(liquid_height_from_bottom, (int, float)):
        if liquid_height_from_bottom + well_core.get_bottom(0).z > location.point.z:
            raise RuntimeError(
                f"{location_check_descriptors.location_type.capitalize()} location {location} is"
                f" inside the liquid in well {well_core.get_display_name()} when it should be outside"
                f"(above) the liquid."
            )
    else:
        # We could raise an error here but that would restrict the use of
        # liquid classes-based transfer to only when LPD is enabled or when liquids are
        # loaded in protocols using `load_liquid`. This can be quite restrictive
        # so we will not raise but just log the details.
        logger.info(
            f"Could not verify height of liquid in well {well_core.get_display_name()}, either"
            f" because the liquid in this well has not been probed or"
            f" liquid was not loaded in this well using `load_liquid` or"
            f" inner geometry is not available for the target well."
            f" Proceeding without verifying if {location_check_descriptors.location_type}"
            f" location is outside the liquid."
        )


def group_wells_for_multi_channel_transfer(
    targets: Sequence[Well],
    nozzle_map: NozzleMapInterface,
    target_name: Literal["source", "destination"],
) -> List[Well]:
    """Takes a list of wells and a nozzle map and returns a list of target wells to address every well given

    This currently only supports 8-tip columns, 12-tip rows and full 96-channel configurations,
    and only is used for 96 and 384 well plates. This assumes the wells are being given in a
    contiguous order (or every other for 384), and will raise if a well is found that does not overlap
    with the first target well given for a sequence, or if not all wells are given for that sequence.
    """
    configuration = nozzle_map.configuration
    active_nozzles = nozzle_map.tip_count

    if (
        (
            (
                configuration == NozzleConfigurationType.COLUMN
                or configuration == NozzleConfigurationType.FULL
            )
            and active_nozzles == 8
        )
        or (configuration == NozzleConfigurationType.ROW and active_nozzles == 12)
        or active_nozzles == 96
    ):
        return _group_wells_for_nozzle_configuration(
            list(targets), nozzle_map, target_name
        )
    else:
        raise ValueError(
            "Unsupported nozzle configuration for well grouping. Set group_wells to False"
            " to only target wells with the primary nozzle for this configuration."
        )


def _group_wells_for_nozzle_configuration(  # noqa: C901
    targets: List[Well],
    nozzle_map: NozzleMapInterface,
    target_name: Literal["source", "destination"],
) -> List[Well]:
    """Groups wells together for a column, row, or full 96 configuration and returns a reduced list of target wells."""
    grouped_wells = []
    active_wells_covered: List[str] = []
    active_labware: Optional[Labware] = None
    alternate_384_well_coverage_count = 0
    labware_format: Optional[str] = None

    # We are assuming the wells are ordered A1, B1, C1... A2, B2, C2..., for columns and
    # A1, A2, A3... B1, B2, B3 for rows. So if the active nozzle is on H row/12 column,
    # reverse the list so the correct primary nozzle is chosen
    reverse_lookup = (
        nozzle_map.starting_nozzle == "H12"
        or (
            nozzle_map.configuration == NozzleConfigurationType.COLUMN
            and nozzle_map.starting_nozzle == "H1"
        )
        or (
            nozzle_map.configuration == NozzleConfigurationType.ROW
            and nozzle_map.starting_nozzle == "A12"
        )
    )
    if reverse_lookup:
        targets.reverse()

    for well in targets:
        # If we have wells that are covered by the pipette's nozzles while primary nozzle is over
        # a target well that aren't accounted for, check if the current well is in that list
        if active_wells_covered:
            if well.parent != active_labware:
                raise ValueError(
                    f"Could not group {target_name} wells to match pipette's nozzle configuration. Ensure that the"
                    " wells are ordered correctly (e.g. rows() for a row layout or columns() for a column layout), or"
                    " set group_wells to False to only target wells with the primary nozzle."
                )

            if well.well_name in active_wells_covered:
                active_wells_covered.remove(well.well_name)
            # If it's a 384 well plate, contiguous wells are not covered by the pipette targeting the
            # initial target well. To support these kinds of transfers given a list of contiguous wells,
            # allow another target well (or up to 4 total for a full 96-tip config) and add those wells
            # to the list of covered wells
            elif labware_format == "384Standard" and (
                alternate_384_well_coverage_count == 0
                or (
                    nozzle_map.tip_count == 96 and alternate_384_well_coverage_count < 3
                )
            ):
                active_wells_covered.extend(
                    list(
                        wells_covered_by_pipette_configuration(
                            nozzle_map,  # type: ignore[arg-type]
                            well.well_name,
                            labware_wells_by_column=[
                                [labware_well.well_name for labware_well in column]
                                for column in well.parent.columns()
                            ],
                        )
                    )
                )
                active_wells_covered.remove(well.well_name)
                grouped_wells.append(well)
                alternate_384_well_coverage_count += 1
            else:
                raise ValueError(
                    f"Could not group {target_name} wells to match pipette's nozzle configuration. Ensure that the"
                    " wells are ordered correctly (e.g. rows() for a row layout or columns() for a column layout), or"
                    " set group_wells to False to only target wells with the primary nozzle."
                )
        # If we have no active wells covered to account for, add a new target well and list of covered wells to check
        else:
            # If the labware is not a 96 or 384 well plate, add this well to the final result and move on to the next
            labware_format = well.parent.parameters["format"]
            if labware_format != "96Standard" and labware_format != "384Standard":
                grouped_wells.append(well)
                continue

            active_wells_covered = list(
                wells_covered_by_pipette_configuration(
                    nozzle_map,  # type: ignore[arg-type]
                    well.well_name,
                    labware_wells_by_column=[
                        [labware_well.well_name for labware_well in column]
                        for column in well.parent.columns()
                    ],
                )
            )
            active_wells_covered.remove(well.well_name)
            grouped_wells.append(well)
            active_labware = well.parent
            alternate_384_well_coverage_count = 0

    if active_wells_covered:
        raise ValueError(
            f"Pipette will access {target_name} wells not provided in the liquid handling command."
            f" Set group_wells to False or include these wells: {active_wells_covered}"
        )

    # If we reversed the lookup of wells, reverse the grouped wells we will return
    if reverse_lookup:
        grouped_wells.reverse()

    return grouped_wells


def check_current_volume_before_dispensing(
    current_volume: float,
    dispense_volume: float,
) -> None:
    """Check if the current volume is valid for dispensing the dispense volume."""
    if current_volume < dispense_volume:
        # Although this should never happen, we can get into an unexpected state
        # following error recovery and not have the expected amount of liquid in the tip.
        # If this happens, we want to raise a useful error so the user can understand
        # the cause of the problem. If we don't make this check for current volume,
        # an unhelpful error might get raised when a '..byVolume' property encounters
        # a negative volume (current_volume - dispense_volume).
        raise RuntimeError(
            f"Cannot dispense {dispense_volume}uL when the tip has only {current_volume}uL."
        )



import logging
from typing import TYPE_CHECKING
from opentrons.types import Point
from opentrons.protocols.execution.types import LoadedLabware, Instruments
from .execute_json_v3 import _get_well

if TYPE_CHECKING:
    from opentrons_shared_data.protocol.types import MoveToWellParams

MODULE_LOG = logging.getLogger(__name__)


def _move_to_well(
    instruments: Instruments, loaded_labware: LoadedLabware, params: "MoveToWellParams"
) -> None:
    pipette_id = params["pipette"]
    pipette = instruments[pipette_id]

    well = _get_well(loaded_labware, params)

    offset = params.get("offset", {})
    offsetPoint = Point(offset.get("x", 0), offset.get("y", 0), offset.get("z", 0))

    pipette.move_to(
        well.bottom().move(offsetPoint),
        force_direct=params.get("forceDirect", False),
        minimum_z_height=params.get("minimumZHeight"),
    )



"""Command models to blow out in place while plunger positions are unknown."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from pydantic import BaseModel

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..pipetting_common import PipetteIdMixin, FlowRateMixin
from ...resources import ensure_ot3_hardware
from ...errors.error_occurrence import ErrorOccurrence
from ...state import update_types

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import Axis

if TYPE_CHECKING:
    from ...execution import PipettingHandler
    from ...state.state import StateView


UnsafeBlowOutInPlaceCommandType = Literal["unsafe/blowOutInPlace"]


class UnsafeBlowOutInPlaceParams(PipetteIdMixin, FlowRateMixin):
    """Payload required to blow-out in place while position is unknown."""

    pass


class UnsafeBlowOutInPlaceResult(BaseModel):
    """Result data from an UnsafeBlowOutInPlace command."""

    pass


class UnsafeBlowOutInPlaceImplementation(
    AbstractCommandImpl[
        UnsafeBlowOutInPlaceParams, SuccessData[UnsafeBlowOutInPlaceResult]
    ]
):
    """UnsafeBlowOutInPlace command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api

    async def execute(
        self, params: UnsafeBlowOutInPlaceParams
    ) -> SuccessData[UnsafeBlowOutInPlaceResult]:
        """Blow-out without moving the pipette even when position is unknown."""
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)
        pipette_location = self._state_view.motion.get_pipette_location(
            params.pipetteId
        )
        await ot3_hardware_api.update_axis_position_estimations(
            [Axis.of_main_tool_actuator(pipette_location.mount.to_hw_mount())]
        )
        await self._pipetting.blow_out_in_place(
            pipette_id=params.pipetteId, flow_rate=params.flowRate
        )
        state_update = update_types.StateUpdate()
        state_update.set_fluid_empty(pipette_id=params.pipetteId)
        state_update.set_pipette_ready_to_aspirate(
            pipette_id=params.pipetteId, ready_to_aspirate=False
        )
        return SuccessData(
            public=UnsafeBlowOutInPlaceResult(), state_update=state_update
        )


class UnsafeBlowOutInPlace(
    BaseCommand[UnsafeBlowOutInPlaceParams, UnsafeBlowOutInPlaceResult, ErrorOccurrence]
):
    """UnsafeBlowOutInPlace command model."""

    commandType: UnsafeBlowOutInPlaceCommandType = "unsafe/blowOutInPlace"
    params: UnsafeBlowOutInPlaceParams
    result: Optional[UnsafeBlowOutInPlaceResult] = None

    _ImplementationCls: Type[
        UnsafeBlowOutInPlaceImplementation
    ] = UnsafeBlowOutInPlaceImplementation


class UnsafeBlowOutInPlaceCreate(BaseCommandCreate[UnsafeBlowOutInPlaceParams]):
    """UnsafeBlowOutInPlace command request model."""

    commandType: UnsafeBlowOutInPlaceCommandType = "unsafe/blowOutInPlace"
    params: UnsafeBlowOutInPlaceParams

    _CommandCls: Type[UnsafeBlowOutInPlace] = UnsafeBlowOutInPlace



from . import (
    models,
    mark_bad_calibration,
)
from .pipette_offset import (
    save_pipette_calibration,
    clear_pipette_offset_calibrations,
    get_pipette_offset,
    delete_pipette_offset_file,
)

from .tip_length import (
    clear_tip_length_calibration,
    create_tip_length_data,
    save_tip_length_calibration,
    tip_lengths_for_pipette,
    load_tip_length_calibration,
    delete_tip_length_calibration,
)

__all__ = [
    "models",
    "mark_bad_calibration",
    "save_pipette_calibration",
    "clear_pipette_offset_calibrations",
    "get_pipette_offset",
    "delete_pipette_offset_file",
    "clear_tip_length_calibration",
    "create_tip_length_data",
    "save_tip_length_calibration",
    "tip_lengths_for_pipette",
    "load_tip_length_calibration",
    "delete_tip_length_calibration",
]



"""Command models for moving any robot axis relative."""

from __future__ import annotations
from typing import Literal, Type, Optional, TYPE_CHECKING, Any

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine.resources import ensure_ot3_hardware

from .common import MotorAxisMapType, DestinationRobotPositionResult

from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
)
from ...errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import GantryMover


MoveAxesRelativeCommandType = Literal["robot/moveAxesRelative"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class MoveAxesRelativeParams(BaseModel):
    """Payload required to move axes relative to position."""

    axis_map: MotorAxisMapType = Field(
        ..., description="A dictionary mapping axes to relative movements in mm."
    )
    speed: float | SkipJsonSchema[None] = Field(
        default=None,
        description="The max velocity to move the axes at. Will fall to hardware defaults if none provided.",
        json_schema_extra=_remove_default,
    )


class MoveAxesRelativeResult(DestinationRobotPositionResult):
    """Result data from the execution of a MoveAxesRelative command."""

    pass


class MoveAxesRelativeImplementation(
    AbstractCommandImpl[MoveAxesRelativeParams, SuccessData[MoveAxesRelativeResult]]
):
    """MoveAxesRelative command implementation."""

    def __init__(
        self,
        gantry_mover: GantryMover,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._gantry_mover = gantry_mover
        self._hardware_api = hardware_api

    async def execute(
        self, params: MoveAxesRelativeParams
    ) -> SuccessData[MoveAxesRelativeResult]:
        """Move the axes on a flex a relative distance."""
        # TODO (lc 08-16-2024) implement `move_axes` for OT 2 hardware controller
        # and then we can remove this validation.
        ensure_ot3_hardware(self._hardware_api)

        current_position = await self._gantry_mover.move_axes(
            axis_map=params.axis_map, speed=params.speed, relative_move=True
        )
        return SuccessData(
            public=MoveAxesRelativeResult(position=current_position),
        )


class MoveAxesRelative(
    BaseCommand[MoveAxesRelativeParams, MoveAxesRelativeResult, ErrorOccurrence]
):
    """MoveAxesRelative command model."""

    commandType: MoveAxesRelativeCommandType = "robot/moveAxesRelative"
    params: MoveAxesRelativeParams
    result: Optional[MoveAxesRelativeResult] = None

    _ImplementationCls: Type[
        MoveAxesRelativeImplementation
    ] = MoveAxesRelativeImplementation


class MoveAxesRelativeCreate(BaseCommandCreate[MoveAxesRelativeParams]):
    """MoveAxesRelative command request model."""

    commandType: MoveAxesRelativeCommandType = "robot/moveAxesRelative"
    params: MoveAxesRelativeParams

    _CommandCls: Type[MoveAxesRelative] = MoveAxesRelative



import logging
import os
import shutil
from enum import Enum
from pathlib import Path
from typing import NamedTuple, Dict, Set

from opentrons_shared_data.robot.types import RobotTypeEnum
from opentrons.config import IS_ROBOT
from opentrons.calibration_storage import (
    delete_robot_deck_attitude,
    gripper_offset,
    ot2,
    ot3,
)


DATA_BOOT_D = Path("/data/boot.d")
AUTHORIZED_KEYS = Path(os.path.expanduser("~/.ssh/authorized_keys"))

log = logging.getLogger(__name__)


class UnrecognizedOption(Exception):
    pass


class CommonResetOption(NamedTuple):
    name: str
    description: str


class ResetOptionId(str, Enum):
    """The available reset options"""

    boot_scripts = "bootScripts"
    deck_calibration = "deckCalibration"
    deck_configuration = "deckConfiguration"
    pipette_offset = "pipetteOffsetCalibrations"
    gripper_offset = "gripperOffsetCalibrations"
    tip_length_calibrations = "tipLengthCalibrations"
    runs_history = "runsHistory"
    on_device_display = "onDeviceDisplay"
    module_calibration = "moduleCalibration"
    authorized_keys = "authorizedKeys"


_OT_2_RESET_OPTIONS = [
    ResetOptionId.boot_scripts,
    ResetOptionId.deck_calibration,
    ResetOptionId.pipette_offset,
    ResetOptionId.tip_length_calibrations,
    ResetOptionId.runs_history,
    ResetOptionId.deck_configuration,
    ResetOptionId.authorized_keys,
]
_FLEX_RESET_OPTIONS = [
    ResetOptionId.boot_scripts,
    ResetOptionId.pipette_offset,
    ResetOptionId.gripper_offset,
    ResetOptionId.runs_history,
    ResetOptionId.on_device_display,
    ResetOptionId.deck_configuration,
    ResetOptionId.module_calibration,
    ResetOptionId.authorized_keys,
]

_settings_reset_options = {
    ResetOptionId.boot_scripts: CommonResetOption(
        name="Boot Scripts", description="Clear custom boot scripts"
    ),
    ResetOptionId.deck_calibration: CommonResetOption(
        name="Deck Calibration",
        description="Clear deck calibration (will also clear pipette offset)",
    ),
    ResetOptionId.pipette_offset: CommonResetOption(
        name="Pipette Offset Calibrations",
        description="Clear pipette offset calibrations",
    ),
    ResetOptionId.gripper_offset: CommonResetOption(
        name="Gripper Offset Calibrations",
        description="Clear gripper offset calibrations",
    ),
    ResetOptionId.tip_length_calibrations: CommonResetOption(
        name="Tip Length Calibrations",
        description="Clear tip length calibrations (will also clear pipette offset)",
    ),
    # TODO(mm, 2022-05-23): runs_history, on_device_display, and deck_configuration are
    # robot-server things, and are not concepts known to this package (the `opentrons` library).
    # This option is defined here only as a convenience for robot-server.
    # Find a way to split things up and define this in robot-server instead.
    ResetOptionId.runs_history: CommonResetOption(
        name="Clear Robot Server Data",
        description=(
            "Erase everything stored by the robot server. This is *not* everything stored"
            " on the robot. It currently includes runs, protocols, labware offsets, and more,"
            " and the exact list may change over time. This is exposed for troubleshooting"
            " and system recovery. The name `runsHistory` is a misnomer, for historical reasons."
        ),
    ),
    ResetOptionId.on_device_display: CommonResetOption(
        name="On-Device Display Configuration",
        description="Clear the configuration of the on-device display (touchscreen)",
    ),
    ResetOptionId.deck_configuration: CommonResetOption(
        name="Deck Configuration",
        description="Clear deck configuration",
    ),
    ResetOptionId.module_calibration: CommonResetOption(
        name="Module Calibrations", description="Clear module offset calibrations"
    ),
    ResetOptionId.authorized_keys: CommonResetOption(
        name="SSH Authorized Keys", description="Clear the ssh authorized keys"
    ),
}


def reset_options(robot_type: RobotTypeEnum) -> Dict[ResetOptionId, CommonResetOption]:
    reset_options_for_robot_type = (
        _OT_2_RESET_OPTIONS if robot_type is RobotTypeEnum.OT2 else _FLEX_RESET_OPTIONS
    )
    return {
        key: _settings_reset_options[key]
        for key in _settings_reset_options
        if key in reset_options_for_robot_type
    }


def reset(options: Set[ResetOptionId], robot_type: RobotTypeEnum) -> None:
    """
    Execute a reset of the requested parts of the user configuration.

    :param options: the parts to reset
    """
    log.info("Reset requested for %s", options)

    if ResetOptionId.boot_scripts in options:
        reset_boot_scripts()

    if ResetOptionId.deck_calibration in options:
        reset_deck_calibration(robot_type)

    if ResetOptionId.pipette_offset in options:
        reset_pipette_offset(robot_type)

    if ResetOptionId.tip_length_calibrations in options:
        reset_tip_length_calibrations(robot_type)

    if ResetOptionId.gripper_offset in options:
        reset_gripper_offset()

    if ResetOptionId.module_calibration in options:
        reset_module_calibration()

    if ResetOptionId.authorized_keys in options:
        reset_authorized_keys()


def reset_boot_scripts() -> None:
    if IS_ROBOT:
        if os.path.exists(DATA_BOOT_D):
            shutil.rmtree(DATA_BOOT_D)
    else:
        log.debug(f"Not on pi, not removing {DATA_BOOT_D}")


def reset_deck_calibration(robot_type: RobotTypeEnum) -> None:
    delete_robot_deck_attitude()
    ot2.clear_pipette_offset_calibrations() if robot_type is RobotTypeEnum.OT2 else ot3.clear_pipette_offset_calibrations()


def reset_pipette_offset(robot_type: RobotTypeEnum) -> None:
    ot2.clear_pipette_offset_calibrations() if robot_type is RobotTypeEnum.OT2 else ot3.clear_pipette_offset_calibrations()


def reset_gripper_offset() -> None:
    gripper_offset.clear_gripper_calibration_offsets()


def reset_tip_length_calibrations(robot_type: RobotTypeEnum) -> None:
    if robot_type is RobotTypeEnum.OT2:
        ot2.clear_tip_length_calibration()
        ot2.clear_pipette_offset_calibrations()
    else:
        raise UnrecognizedOption("Tip Length calibration not supported on Flex")


def reset_module_calibration() -> None:
    try:
        from opentrons.calibration_storage.ot3.module_offset import (
            clear_module_offset_calibrations,
        )

        clear_module_offset_calibrations()
    except ImportError:
        log.warning("Tried to clear module offset calibrations on an OT-2")


def reset_authorized_keys() -> None:
    if IS_ROBOT and os.path.exists(AUTHORIZED_KEYS):
        with open(AUTHORIZED_KEYS, "w") as fh:
            fh.write("")



"""Support for create_protocol_engine module."""
from . import ProtocolEngine
from ..hardware_control import HardwareControlAPI

from opentrons.protocol_runner import protocol_runner, RunOrchestrator


def create_run_orchestrator(
    hardware_api: HardwareControlAPI,
    protocol_engine: ProtocolEngine,
) -> RunOrchestrator:
    """Create a RunOrchestrator instance."""
    return RunOrchestrator(
        protocol_engine=protocol_engine,
        hardware_api=hardware_api,
        setup_runner=protocol_runner.LiveRunner(
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
        ),
        fixit_runner=protocol_runner.LiveRunner(
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
        ),
        protocol_live_runner=protocol_runner.LiveRunner(
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
        ),
    )



"""ProtocolEngine-based Well core implementations."""
from typing import Optional, Union

from opentrons_shared_data.labware.constants import WELL_NAME_PATTERN

from opentrons.types import Point, Mount, MountType

from opentrons.protocol_engine import WellLocation, WellOrigin, WellOffset
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocols.api_support.util import UnsupportedAPIError
from opentrons.protocol_engine.types.liquid_level_detection import (
    SimulatedProbeResult,
    LiquidTrackingType,
)
from opentrons.protocol_engine.errors import PipetteNotAttachedError

from . import point_calculations
from . import stringify
from ..well import AbstractWellCore
from ..._liquid import Liquid


class WellCore(AbstractWellCore):
    """Well API core using a ProtocolEngine.

    Args:
        name: The well's name in the labware, e.g. `A1`.
        labware_id: The ProtocolEngine ID of the well's parent labware.
        engine_client: Synchronous ProtocolEngine client.
    """

    def __init__(self, name: str, labware_id: str, engine_client: EngineClient) -> None:
        self._labware_id = labware_id
        self._engine_client = engine_client
        self._definition = engine_client.state.labware.get_well_definition(
            labware_id=labware_id, well_name=name
        )

        name_match = WELL_NAME_PATTERN.match(name)
        self._name = name
        self._row_name = name_match.group(1) if name_match is not None else ""
        self._column_name = name_match.group(2) if name_match is not None else ""

    @property
    def labware_id(self) -> str:
        """Get the ID of the well's parent labware."""
        return self._labware_id

    @property
    def diameter(self) -> Optional[float]:
        """Get the well's diameter, if circular."""
        return (
            self._definition.diameter if self._definition.shape == "circular" else None
        )

    @property
    def length(self) -> Optional[float]:
        """Get the well's length, if rectangular."""
        return (
            self._definition.xDimension
            if self._definition.shape == "rectangular"
            else None
        )

    @property
    def width(self) -> Optional[float]:
        """Get the well's width, if rectangular."""
        return (
            self._definition.yDimension
            if self._definition.shape == "rectangular"
            else None
        )

    @property
    def depth(self) -> float:
        """Get the well's depth."""
        return self._definition.depth

    def has_tip(self) -> bool:
        """Whether the well contains a tip."""
        return self._engine_client.state.tips.has_clean_tip(
            self._labware_id, self._name
        )

    def set_has_tip(self, value: bool) -> None:
        """Set the well as containing or not containing a tip."""
        raise UnsupportedAPIError(
            api_element="Manually setting the tip state of a well in a tip rack",
        )

    def get_display_name(self) -> str:
        """Get the full display name of the well (e.g. "A1 of Some Labware on 5")."""
        return stringify.well(
            engine_client=self._engine_client,
            well_name=self._name,
            labware_id=self._labware_id,
        )

    def get_name(self) -> str:
        """Get the name of the well (e.g. "A1")."""
        return self._name

    def get_column_name(self) -> str:
        """Get the column portion of the well name (e.g. "1")."""
        return self._column_name

    def get_row_name(self) -> str:
        """Get the row portion of the well name (e.g. "A")."""
        return self._row_name

    def get_max_volume(self) -> float:
        """Get the well's maximum liquid volume."""
        return self._definition.totalLiquidVolume

    def get_top(self, z_offset: float) -> Point:
        """Get the coordinate of the well's top, with a z-offset."""
        return self._engine_client.state.geometry.get_well_position(
            well_name=self._name,
            labware_id=self._labware_id,
            well_location=WellLocation(
                origin=WellOrigin.TOP, offset=WellOffset(x=0, y=0, z=z_offset)
            ),
        )

    def get_bottom(self, z_offset: float) -> Point:
        """Get the coordinate of the well's bottom, with a z-offset."""
        return self._engine_client.state.geometry.get_well_position(
            well_name=self._name,
            labware_id=self._labware_id,
            well_location=WellLocation(
                origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=z_offset)
            ),
        )

    def get_center(self) -> Point:
        """Get the coordinate of the well's center."""
        return self._engine_client.state.geometry.get_well_position(
            well_name=self._name,
            labware_id=self._labware_id,
            well_location=WellLocation(origin=WellOrigin.CENTER),
        )

    def get_meniscus(self) -> Union[Point, SimulatedProbeResult]:
        """Get the coordinate of the well's meniscus."""
        current_liquid_height = self.current_liquid_height()
        if isinstance(current_liquid_height, float):
            return self.get_bottom(z_offset=current_liquid_height)
        else:
            return current_liquid_height

    def load_liquid(
        self,
        liquid: Liquid,
        volume: float,
    ) -> None:
        """Load liquid into a well.

        If the well is known to be empty, use ``load_empty()`` instead of calling this with a 0.0 volume.
        """
        self._engine_client.execute_command(
            cmd.LoadLiquidParams(
                labwareId=self._labware_id,
                liquidId=liquid._id,
                volumeByWell={self._name: volume},
            )
        )

    def from_center_cartesian(self, x: float, y: float, z: float) -> Point:
        """Gets point in deck coordinates based on percentage of the radius of each axis."""
        well_size = self._engine_client.state.labware.get_well_size(
            labware_id=self.labware_id, well_name=self._name
        )

        return point_calculations.get_relative_offset(
            point=self.get_center(),
            size=well_size,
            x_ratio=x,
            y_ratio=y,
            z_ratio=z,
        )

    def estimate_liquid_height_after_pipetting(
        self,
        mount: Mount | str,
        operation_volume: float,
    ) -> LiquidTrackingType:
        """Return an estimate of liquid height after pipetting without raising an error."""
        labware_id = self.labware_id
        well_name = self._name
        if isinstance(mount, Mount):
            mount_type = MountType.from_hw_mount(mount)
        else:
            mount_type = MountType(mount)
        pipette_from_mount = self._engine_client.state.pipettes.get_by_mount(mount_type)
        if pipette_from_mount is None:
            raise PipetteNotAttachedError(f"No pipette present on mount {mount}")
        pipette_id = pipette_from_mount.id
        starting_liquid_height = self.current_liquid_height()
        projected_final_height = (
            self._engine_client.state.geometry.get_well_height_after_liquid_handling(
                labware_id=labware_id,
                well_name=well_name,
                pipette_id=pipette_id,
                initial_height=starting_liquid_height,
                volume=operation_volume,
            )
        )
        return projected_final_height

    def current_liquid_height(self) -> LiquidTrackingType:
        """Return the current liquid height within a well."""
        labware_id = self.labware_id
        well_name = self._name
        return self._engine_client.state.geometry.get_meniscus_height(
            labware_id=labware_id, well_name=well_name
        )

    def get_liquid_volume(self) -> LiquidTrackingType:
        """Return the current volume in a well."""
        labware_id = self.labware_id
        well_name = self._name
        return self._engine_client.state.geometry.get_current_well_volume(
            labware_id=labware_id, well_name=well_name
        )

    def height_from_volume(self, volume: LiquidTrackingType) -> LiquidTrackingType:
        """Return the height in a well corresponding to a given volume."""
        labware_id = self.labware_id
        well_name = self._name
        return self._engine_client.state.geometry.get_well_height_at_volume(
            labware_id=labware_id, well_name=well_name, volume=volume
        )

    def volume_from_height(self, height: LiquidTrackingType) -> LiquidTrackingType:
        """Return the volume contained in a well at any height."""
        labware_id = self.labware_id
        well_name = self._name
        return self._engine_client.state.geometry.get_well_volume_at_height(
            labware_id=labware_id, well_name=well_name, height=height
        )
